# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for gkeonprem API clients for Bare Metal Standalone cluster resources."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import encoding
from apitools.base.py import list_pager
from googlecloudsdk.api_lib.container.gkeonprem import client
from googlecloudsdk.api_lib.container.gkeonprem import update_mask
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.calliope import parser_extensions
from googlecloudsdk.core import properties
from googlecloudsdk.generated_clients.apis.gkeonprem.v1 import gkeonprem_v1_messages as messages


class _BareMetalStandaloneClusterClient(client.ClientBase):
  """Base class for GKE OnPrem Bare Metal Standalone API clients."""

  def _sr_iov_config_enabled(self, args: parser_extensions.Namespace):
    if 'enable_sr_iov_config' in args.GetSpecifiedArgsDict():
      return True
    elif 'disable_sr_iov_config' in args.GetSpecifiedArgsDict():
      return False
    else:
      return None

  def _sr_iov_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneSrIovConfig."""
    kwargs = {
        'enabled': self._sr_iov_config_enabled(args),
    }

    if self.IsSet(kwargs):
      return messages.BareMetalStandaloneSrIovConfig(**kwargs)

    return None

  def _island_mode_cidr_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneIslandModeCidrConfig."""
    kwargs = {
        'serviceAddressCidrBlocks': getattr(
            args, 'island_mode_service_address_cidr_blocks', []
        ),
    }

    if any(kwargs.values()):
      return messages.BareMetalStandaloneIslandModeCidrConfig(**kwargs)

    return None

  def _network_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneNetworkConfig."""
    kwargs = {
        'islandModeCidr': self._island_mode_cidr_config(args),
        'srIovConfig': self._sr_iov_config(args),
    }

    if any(kwargs.values()):
      return messages.BareMetalStandaloneNetworkConfig(**kwargs)

    return None

  def _address_pools_from_file(self, args: parser_extensions.Namespace):
    """Constructs proto message field address_pools."""
    if not args.metal_lb_address_pools_from_file:
      return []

    address_pools = args.metal_lb_address_pools_from_file.get(
        'addressPools', []
    )

    if not address_pools:
      self._raise_bad_argument_exception_error(
          '--metal_lb_address_pools_from_file',
          'addressPools',
          'Metal LB address pools file',
      )

    address_pool_messages = [
        self._metal_lb_address_pool(address_pool)
        for address_pool in address_pools
    ]

    return address_pool_messages

  def _metal_lb_address_pool(self, address_pool):
    """Constructs proto message BareMetalStandaloneLoadBalancerAddressPool."""
    addresses = address_pool.get('addresses', [])
    if not addresses:
      self._raise_bad_argument_exception_error(
          '--metal_lb_address_pools_from_file',
          'addresses',
          'Metal LB address pools file',
      )

    pool = address_pool.get('pool', None)
    if not pool:
      self._raise_bad_argument_exception_error(
          '--metal_lb_address_pools_from_file',
          'pool',
          'Metal LB address pools file',
      )

    kwargs = {
        'addresses': addresses,
        'avoidBuggyIps': address_pool.get('avoidBuggyIPs', None),
        'manualAssign': address_pool.get('manualAssign', None),
        'pool': pool,
    }

    return messages.BareMetalStandaloneLoadBalancerAddressPool(**kwargs)

  def _address_pools_from_flag(self, args: parser_extensions.Namespace):
    if not args.metal_lb_address_pools:
      return []

    address_pools = []
    for address_pool in args.metal_lb_address_pools:
      address_pools.append(
          messages.BareMetalStandaloneLoadBalancerAddressPool(
              addresses=address_pool.get('addresses', []),
              avoidBuggyIps=address_pool.get('avoid-buggy-ips', None),
              manualAssign=address_pool.get('manual-assign', None),
              pool=address_pool.get('pool', None),
          )
      )

    return address_pools

  def _metal_lb_node_config(self, metal_lb_node_config):
    """Constructs proto message BareMetalNodeConfig."""
    node_ip = metal_lb_node_config.get('nodeIP', '')
    if not node_ip:
      self._raise_bad_argument_exception_error(
          '--metal_lb_load_balancer_node_configs_from_file',
          'nodeIp',
          'Metal LB Node configs file',
      )

    kwargs = {
        'nodeIp': node_ip,
        'labels': self._node_labels(metal_lb_node_config.get('labels', {})),
    }

    return messages.BareMetalNodeConfig(**kwargs)

  def _metal_lb_node_configs_from_file(self, args: parser_extensions.Namespace):
    """Constructs proto message field node_configs."""
    if not args.metal_lb_load_balancer_node_configs_from_file:
      return []

    metal_lb_node_configs = (
        args.metal_lb_load_balancer_node_configs_from_file.get(
            'nodeConfigs', []
        )
    )

    if not metal_lb_node_configs:
      self._raise_bad_argument_exception_error(
          '--metal_lb_load_balancer_node_configs_from_file',
          'nodeConfigs',
          'Metal LB Node configs file',
      )

    metal_lb_node_configs_messages = []
    for metal_lb_node_config in metal_lb_node_configs:
      metal_lb_node_configs_messages.append(
          self._metal_lb_node_config(metal_lb_node_config)
      )

    return metal_lb_node_configs_messages

  def parse_node_labels(self, node_labels):
    """Validates and parses a node label object.

    Args:
      node_labels: str of key-val pairs separated by ';' delimiter.

    Returns:
      If label is valid, returns a dict mapping message LabelsValue to its
      value, otherwise, raise ArgumentTypeError.
      For example,
      {
          'key': LABEL_KEY
          'value': LABEL_VALUE
      }
    """
    if not node_labels.get('labels'):
      return None

    input_node_labels = node_labels.get('labels', '').split(';')
    additional_property_messages = []

    for label in input_node_labels:
      key_val_pair = label.split('=')
      if len(key_val_pair) != 2:
        raise arg_parsers.ArgumentTypeError(
            'Node Label [{}] not in correct format, expect KEY=VALUE.'.format(
                input_node_labels
            )
        )
      additional_property_messages.append(
          messages.BareMetalNodeConfig.LabelsValue.AdditionalProperty(
              key=key_val_pair[0], value=key_val_pair[1]
          )
      )

    labels_value_message = messages.BareMetalNodeConfig.LabelsValue(
        additionalProperties=additional_property_messages
    )

    return labels_value_message

  def node_config(self, node_config_args):
    """Constructs proto message BareMetalNodeConfig."""
    kwargs = {
        'nodeIp': node_config_args.get('node-ip', ''),
        'labels': self.parse_node_labels(node_config_args),
    }

    return self._set_config_if_exists(messages.BareMetalNodeConfig, kwargs)

  def _metal_lb_node_configs_from_flag(self, args: parser_extensions.Namespace):
    """Constructs proto message field node_configs."""
    node_config_flag_value = (
        getattr(args, 'metal_lb_load_balancer_node_configs', [])
        if args.metal_lb_load_balancer_node_configs
        else []
    )

    return [
        self.node_config(node_config) for node_config in node_config_flag_value
    ]

  def _metal_lb_node_taints(self, args: parser_extensions.Namespace):
    """Constructs proto message NodeTaint."""
    taint_messages = []
    node_taints = getattr(args, 'metal_lb_load_balancer_node_taints', {})
    if not node_taints:
      return []

    for node_taint in node_taints.items():
      taint_object = self._parse_node_taint(node_taint)
      taint_messages.append(messages.NodeTaint(**taint_object))

    return taint_messages

  def _metal_lb_labels(self, args: parser_extensions.Namespace):
    """Constructs proto message LabelsValue."""
    node_labels = getattr(args, 'metal_lb_load_balancer_node_labels', {})
    additional_property_messages = []

    if not node_labels:
      return None

    for key, value in node_labels.items():
      additional_property_messages.append(
          messages.BareMetalNodePoolConfig.LabelsValue.AdditionalProperty(
              key=key, value=value
          )
      )

    labels_value_message = messages.BareMetalNodePoolConfig.LabelsValue(
        additionalProperties=additional_property_messages
    )

    return labels_value_message

  def _metal_lb_load_balancer_node_pool_config(
      self, args: parser_extensions.Namespace
  ):
    """Constructs proto message BareMetalNodePoolConfig."""
    if (
        'metal_lb_load_balancer_node_configs_from_file'
        in args.GetSpecifiedArgsDict()
    ):
      metal_lb_node_configs = self._metal_lb_node_configs_from_file(args)
    else:
      metal_lb_node_configs = self._metal_lb_node_configs_from_flag(args)

    kwargs = {
        'nodeConfigs': metal_lb_node_configs,
        'labels': self._metal_lb_labels(args),
        'taints': self._metal_lb_node_taints(args),
    }

    if any(kwargs.values()):
      return messages.BareMetalNodePoolConfig(**kwargs)

    return None

  def _metal_lb_node_pool_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneLoadBalancerNodePoolConfig."""
    kwargs = {
        'nodePoolConfig': self._metal_lb_load_balancer_node_pool_config(args),
    }

    if any(kwargs.values()):
      return messages.BareMetalStandaloneLoadBalancerNodePoolConfig(**kwargs)

    return None

  def _metal_lb_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneMetalLbConfig."""
    if 'metal_lb_address_pools_from_file' in args.GetSpecifiedArgsDict():
      address_pools = self._address_pools_from_file(args)
    else:
      address_pools = self._address_pools_from_flag(args)
    kwargs = {
        'addressPools': address_pools,
        'loadBalancerNodePoolConfig': self._metal_lb_node_pool_config(args),
    }

    if any(kwargs.values()):
      return messages.BareMetalStandaloneMetalLbConfig(**kwargs)

    return None

  def _bgp_address_pools_from_file(self, args: parser_extensions.Namespace):
    """Constructs proto message field address_pools."""
    if not args.bgp_lb_address_pools_from_file:
      return []

    address_pools = args.bgp_lb_address_pools_from_file.get(
        'addressPools', []
    )

    if not address_pools:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_address_pools_from_file',
          'addressPools',
          'BGP LB address pools file',
      )

    address_pool_messages = []
    for address_pool in address_pools:
      address_pool_messages.append(self._bgp_lb_address_pool(address_pool))

    return address_pool_messages

  def _bgp_lb_address_pool(self, address_pool):
    """Constructs proto message BareMetalStandaloneLoadBalancerAddressPool."""
    addresses = address_pool.get('addresses', [])
    if not addresses:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_address_pools_from_file',
          'addresses',
          'BGP LB address pools file',
      )

    pool = address_pool.get('pool', None)
    if not pool:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_address_pools_from_file',
          'pool',
          'BGP LB address pools file',
      )

    kwargs = {
        'addresses': addresses,
        'avoidBuggyIps': address_pool.get('avoidBuggyIPs', None),
        'manualAssign': address_pool.get('manualAssign', None),
        'pool': pool,
    }

    return messages.BareMetalStandaloneLoadBalancerAddressPool(**kwargs)

  def _bgp_address_pools_from_flag(self, args: parser_extensions.Namespace):
    if not args.bgp_lb_address_pools:
      return []

    address_pools = []
    for address_pool in args.bgp_lb_address_pools:
      address_pools.append(
          messages.BareMetalStandaloneLoadBalancerAddressPool(
              addresses=address_pool.get('addresses', []),
              avoidBuggyIps=address_pool.get('avoid-buggy-ips', None),
              manualAssign=address_pool.get('manual-assign', None),
              pool=address_pool.get('pool', None),
          )
      )

    return address_pools

  def _bgp_lb_node_config(self, bgp_lb_node_config):
    """Constructs proto message BareMetalNodeConfig."""
    node_ip = bgp_lb_node_config.get('nodeIP', '')
    if not node_ip:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_load_balancer_node_configs_from_file',
          'nodeIP',
          'BGP LB Node configs file',
      )

    kwargs = {
        'nodeIp': node_ip,
        'labels': self._node_labels(bgp_lb_node_config.get('labels', {})),
    }

    return messages.BareMetalNodeConfig(**kwargs)

  def _bgp_lb_node_configs_from_file(self, args: parser_extensions.Namespace):
    """Constructs proto message field node_configs."""
    if not args.bgp_lb_load_balancer_node_configs_from_file:
      return []

    bgp_lb_node_configs = (
        args.bgp_lb_load_balancer_node_configs_from_file.get(
            'nodeConfigs', []
        )
    )

    if not bgp_lb_node_configs:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_load_balancer_node_configs_from_file',
          'nodeConfigs',
          'BGP LB Node configs file',
      )

    bgp_lb_node_configs_messages = []
    for bgp_lb_node_config in bgp_lb_node_configs:
      bgp_lb_node_configs_messages.append(
          self._bgp_lb_node_config(bgp_lb_node_config)
      )

    return bgp_lb_node_configs_messages

  def _bgp_lb_node_configs_from_flag(self, args: parser_extensions.Namespace):
    """Constructs proto message field node_configs."""
    node_config_flag_value = (
        getattr(args, 'bgp_lb_load_balancer_node_configs', [])
        if args.bgp_lb_load_balancer_node_configs
        else []
    )

    return [
        self.node_config(node_config) for node_config in node_config_flag_value
    ]

  def _bgp_lb_node_taints(self, args: parser_extensions.Namespace):
    """Constructs proto message NodeTaint."""
    taint_messages = []
    node_taints = getattr(args, 'bgp_lb_load_balancer_node_taints', {})
    if not node_taints:
      return []

    for node_taint in node_taints.items():
      taint_object = self._parse_node_taint(node_taint)
      taint_messages.append(messages.NodeTaint(**taint_object))

    return taint_messages

  def _bgp_lb_labels(self, args: parser_extensions.Namespace):
    """Constructs proto message LabelsValue."""
    node_labels = getattr(args, 'bgp_lb_load_balancer_node_labels', {})
    additional_property_messages = []

    if not node_labels:
      return None

    for key, value in node_labels.items():
      additional_property_messages.append(
          messages.BareMetalNodePoolConfig.LabelsValue.AdditionalProperty(
              key=key, value=value
          )
      )

    labels_value_message = messages.BareMetalNodePoolConfig.LabelsValue(
        additionalProperties=additional_property_messages
    )

    return labels_value_message

  def _bgp_lb_load_balancer_node_pool_config(
      self, args: parser_extensions.Namespace
  ):
    """Constructs proto message BareMetalNodePoolConfig."""
    if (
        'bgp_lb_load_balancer_node_configs_from_file'
        in args.GetSpecifiedArgsDict()
    ):
      bgp_lb_node_configs = self._bgp_lb_node_configs_from_file(args)
    else:
      bgp_lb_node_configs = self._bgp_lb_node_configs_from_flag(args)

    kwargs = {
        'nodeConfigs': bgp_lb_node_configs,
        'labels': self._bgp_lb_labels(args),
        'taints': self._bgp_lb_node_taints(args),
    }

    if any(kwargs.values()):
      return messages.BareMetalNodePoolConfig(**kwargs)

    return None

  def _bgp_lb_node_pool_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneLoadBalancerNodePoolConfig."""
    kwargs = {
        'nodePoolConfig': self._bgp_lb_load_balancer_node_pool_config(args),
    }

    if any(kwargs.values()):
      return messages.BareMetalStandaloneLoadBalancerNodePoolConfig(**kwargs)

    return None

  def _bgp_peer_configs_from_file(self, args: parser_extensions.Namespace):
    """Constructs proto message field address_pools."""
    if not args.bgp_lb_peer_configs_from_file:
      return []

    peer_configs = args.bgp_lb_peer_configs_from_file.get('bgpPeerConfigs', [])

    if not peer_configs:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_peer_configs_from_file',
          'bgpPeerConfigs',
          'BGP LB peer configs file',
      )

    peer_configs_messages = []
    for peer_config in peer_configs:
      peer_configs_messages.append(self._peer_configs(peer_config))

    return peer_configs_messages

  def _peer_configs(self, peer_config):
    """Constructs proto message BareMetalStandaloneBgpPeerConfig."""
    asn = peer_config.get('asn', None)
    if not asn:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_peer_configs_from_file',
          'asn',
          'BGP LB peer configs file',
      )

    ip_address = peer_config.get('ipAddress', None)
    if not ip_address:
      self._raise_bad_argument_exception_error(
          '--bgp_lb_peer_configs_from_file',
          'ipAddress',
          'BGP LB peer configs file',
      )

    kwargs = {
        'asn': asn,
        'ipAddress': ip_address,
        'controlPlaneNodes': peer_config.get('controlPlaneNodes', []),
    }

    return messages.BareMetalStandaloneBgpPeerConfig(**kwargs)

  def _bgp_peer_configs_from_flag(self, args: parser_extensions.Namespace):
    if not args.bgp_lb_peer_configs:
      return []

    peer_configs = []
    for peer_config in args.bgp_lb_peer_configs:
      peer_configs.append(
          messages.BareMetalStandaloneBgpPeerConfig(
              controlPlaneNodes=peer_config.get('control-plane-nodes', []),
              asn=peer_config.get('asn', None),
              ipAddress=peer_config.get('ip-address', None),
          )
      )

    return peer_configs

  def _bgp_lb_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneBgpLbConfig."""
    if 'bgp_lb_address_pools_from_file' in args.GetSpecifiedArgsDict():
      address_pools = self._bgp_address_pools_from_file(args)
    else:
      address_pools = self._bgp_address_pools_from_flag(args)

    if 'bgp_lb_peer_configs_from_file' in args.GetSpecifiedArgsDict():
      peer_configs = self._bgp_peer_configs_from_file(args)
    else:
      peer_configs = self._bgp_peer_configs_from_flag(args)

    kwargs = {
        'addressPools': address_pools,
        'asn': getattr(args, 'bgp_lb_asn', None),
        'loadBalancerNodePoolConfig': self._bgp_lb_node_pool_config(args),
        'bgpPeerConfigs': peer_configs,
    }

    if any(kwargs.values()):
      return messages.BareMetalStandaloneBgpLbConfig(**kwargs)

    return None

  def _load_balancer_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneLoadBalancerConfig."""
    kwargs = {
        'metalLbConfig': self._metal_lb_config(args),
        'bgpLbConfig': self._bgp_lb_config(args),
    }

    if any(kwargs.values()):
      return messages.BareMetalStandaloneLoadBalancerConfig(**kwargs)

    return None

  def _node_labels(self, labels):
    """Constructs proto message LabelsValue."""
    additional_property_messages = []
    if not labels:
      return None

    for key, value in labels.items():
      additional_property_messages.append(
          messages.BareMetalNodeConfig.LabelsValue.AdditionalProperty(
              key=key, value=value
          )
      )

    labels_value_message = messages.BareMetalNodeConfig.LabelsValue(
        additionalProperties=additional_property_messages
    )

    return labels_value_message

  def _control_plane_node_config(self, control_plane_node_config):
    """Constructs proto message BareMetalNodeConfig."""
    node_ip = control_plane_node_config.get('nodeIP', '')
    if not node_ip:
      self._raise_bad_argument_exception_error(
          '--control_plane_node_configs_from_file',
          'nodeIP',
          'Control Plane Node configs file',
      )

    kwargs = {
        'nodeIp': node_ip,
        'labels': self._node_labels(
            control_plane_node_config.get('labels', {})
        ),
    }

    return messages.BareMetalNodeConfig(**kwargs)

  def _control_plane_node_configs_from_file(
      self, args: parser_extensions.Namespace
  ):
    """Constructs proto message field node_configs."""
    if not args.control_plane_node_configs_from_file:
      return []

    control_plane_node_configs = args.control_plane_node_configs_from_file.get(
        'nodeConfigs', []
    )

    if not control_plane_node_configs:
      self._raise_bad_argument_exception_error(
          '--control_plane_node_configs_from_file',
          'nodeConfigs',
          'Control Plane Node configs file',
      )

    control_plane_node_configs_messages = []
    for control_plane_node_config in control_plane_node_configs:
      control_plane_node_configs_messages.append(
          self._control_plane_node_config(control_plane_node_config)
      )

    return control_plane_node_configs_messages

  def _raise_bad_argument_exception_error(self, flag, field, file):
    raise exceptions.BadArgumentException(
        flag, 'Missing field [' + field + '] in ' + file
    )

  def _control_plane_node_configs_from_flag(
      self, args: parser_extensions.Namespace
  ):
    """Constructs proto message field node_configs."""
    node_configs = []
    node_config_flag_value = getattr(args, 'control_plane_node_configs', None)
    if node_config_flag_value:
      for node_config in node_config_flag_value:
        node_configs.append(self.node_config(node_config))

    return node_configs

  def _control_plane_node_taints(self, args: parser_extensions.Namespace):
    """Constructs proto message NodeTaint."""
    taint_messages = []
    node_taints = getattr(args, 'control_plane_node_taints', {})
    if not node_taints:
      return []

    for node_taint in node_taints.items():
      taint_object = self._parse_node_taint(node_taint)
      taint_messages.append(messages.NodeTaint(**taint_object))

    return taint_messages

  def _control_plane_node_labels(self, args: parser_extensions.Namespace):
    """Constructs proto message LabelsValue."""
    node_labels = getattr(args, 'control_plane_node_labels', {})
    additional_property_messages = []
    if not node_labels:
      return None

    for key, value in node_labels.items():
      additional_property_messages.append(
          messages.BareMetalNodePoolConfig.LabelsValue.AdditionalProperty(
              key=key, value=value
          )
      )

    labels_value_message = messages.BareMetalNodePoolConfig.LabelsValue(
        additionalProperties=additional_property_messages
    )

    return labels_value_message

  def _node_pool_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalNodePoolConfig."""
    if 'control_plane_node_configs_from_file' in args.GetSpecifiedArgsDict():
      node_configs = self._control_plane_node_configs_from_file(args)
    else:
      node_configs = self._control_plane_node_configs_from_flag(args)

    kwargs = {
        'nodeConfigs': node_configs,
        'labels': self._control_plane_node_labels(args),
        'taints': self._control_plane_node_taints(args),
    }

    return self._set_config_if_exists(messages.BareMetalNodePoolConfig, kwargs)

  def _control_plane_node_pool_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneControlPlaneNodePoolConfig."""
    kwargs = {
        'nodePoolConfig': self._node_pool_config(args),
    }

    return self._set_config_if_exists(
        messages.BareMetalStandaloneControlPlaneNodePoolConfig, kwargs
    )

  def _api_server_args(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneApiServerArgument."""
    api_server_args = []
    api_server_args_flag_value = getattr(args, 'api_server_args', None)
    if api_server_args_flag_value:
      for key, val in api_server_args_flag_value.items():
        api_server_args.append(
            messages.BareMetalStandaloneApiServerArgument(
                argument=key, value=val
            )
        )

    return api_server_args

  def _control_plane_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneControlPlaneConfig."""
    kwargs = {
        'controlPlaneNodePoolConfig': self._control_plane_node_pool_config(
            args
        ),
        'apiServerArgs': self._api_server_args(args),
    }

    return self._set_config_if_exists(
        messages.BareMetalStandaloneControlPlaneConfig, kwargs
    )

  def _cluster_operations_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneClusterOperationsConfig."""
    kwargs = {
        'enableApplicationLogs': getattr(args, 'enable_application_logs', None),
    }

    return self._set_config_if_exists(
        messages.BareMetalStandaloneClusterOperationsConfig, kwargs
    )

  def _maintenance_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneMaintenanceConfig."""
    kwargs = {
        'maintenanceAddressCidrBlocks': getattr(
            args, 'maintenance_address_cidr_blocks', []
        ),
    }

    return self._set_config_if_exists(
        messages.BareMetalStandaloneMaintenanceConfig, kwargs
    )

  # TODO(b/257292798): Move to common directory
  def _cluster_users(self, args: parser_extensions.Namespace):
    """Constructs repeated proto message ClusterUser."""
    cluster_user_messages = []
    admin_users = getattr(args, 'admin_users', None)
    if admin_users:
      return [
          messages.ClusterUser(username=admin_user)
          for admin_user in admin_users
      ]

    # On update, skip setting default value.
    if args.command_path[-1] == 'update':
      return None

    # On create, client side default admin user to the current gcloud user.
    gcloud_config_core_account = properties.VALUES.core.account.Get()
    if gcloud_config_core_account:
      default_admin_user_message = messages.ClusterUser(
          username=gcloud_config_core_account
      )
      return cluster_user_messages.append(default_admin_user_message)

    return None

  def _authorization(self, args: parser_extensions.Namespace):
    """Constructs proto message Authorization."""
    kwargs = {
        'adminUsers': self._cluster_users(args),
    }

    return self._set_config_if_exists(messages.Authorization, kwargs)

  def _security_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneSecurityConfig."""
    kwargs = {
        'authorization': self._authorization(args),
    }

    return self._set_config_if_exists(
        messages.BareMetalStandaloneSecurityConfig, kwargs
    )

  def _node_access_config(self, args: parser_extensions.Namespace):
    """Constructs proto message BareMetalStandaloneNodeAccessConfig."""
    kwargs = {
        'loginUser': getattr(args, 'login_user', 'root'),
    }

    return self._set_config_if_exists(
        messages.BareMetalStandaloneNodeAccessConfig, kwargs
    )

  def _add_annotations(self, args: parser_extensions.Namespace):
    """Constructs proto message AnnotationsValue for adding new annotations."""
    curr_annotations = self._get_current_annotations(args)
    for key, value in args.add_annotations.items():
      curr_annotations[key] = value

    return self._dict_to_annotations_message(curr_annotations)

  def _clear_annotations(self, args: parser_extensions.Namespace):
    """Constructs proto message AnnotationsValue for clearing annotations."""
    return messages.BareMetalStandaloneCluster.AnnotationsValue()

  def _remove_annotations(self, args: parser_extensions.Namespace):
    """Constructs proto message AnnotationsValue for removing annotations."""
    curr_annotations = self._get_current_annotations(args)
    updated_annotations = {
        key: value
        for key, value in curr_annotations.items()
        if key not in args.remove_annotations
    }

    return self._dict_to_annotations_message(updated_annotations)

  def _set_annotations(self, args: parser_extensions.Namespace):
    """Constructs proto message AnnotationsValue for setting annotations."""
    return self._dict_to_annotations_message(args.set_annotations)

  def _get_current_annotations(self, args: parser_extensions.Namespace):
    """Fetches the standalone cluster annotations."""
    cluster_ref = args.CONCEPTS.standalone_cluster.Parse()
    cluster_response = self.Describe(cluster_ref)

    curr_annotations = {}
    if cluster_response.annotations:
      for annotation in cluster_response.annotations.additionalProperties:
        curr_annotations[annotation.key] = annotation.value

    return curr_annotations

  def _update_annotations(self, args: parser_extensions.Namespace):
    """Constructs proto message AnnotationsValue for update command."""
    if 'add_annotations' in args.GetSpecifiedArgsDict():
      return self._add_annotations(args)

    if 'clear_annotations' in args.GetSpecifiedArgsDict():
      return self._clear_annotations(args)

    if 'remove_annotations' in args.GetSpecifiedArgsDict():
      return self._remove_annotations(args)

    if 'set_annotations' in args.GetSpecifiedArgsDict():
      return self._set_annotations(args)

    return None

  def _dict_to_annotations_message(self, annotations):
    """Converts key-val pairs to proto message AnnotationsValue."""
    additional_property_messages = []
    if not annotations:
      return None

    for key, value in annotations.items():
      additional_property_messages.append(
          (
              messages.BareMetalStandaloneCluster.AnnotationsValue.AdditionalProperty
          )(key=key, value=value)
      )

    annotation_value_message = (
        messages.BareMetalStandaloneCluster.AnnotationsValue(
            additionalProperties=additional_property_messages
        )
    )
    return annotation_value_message

  def _annotations(self, args: parser_extensions.Namespace):
    """Constructs proto message AnnotationsValue."""
    if args.command_path[-1] == 'update':
      return self._update_annotations(args)

    return None

  def _bare_metal_standalone_cluster(self, args: parser_extensions.Namespace):
    """Constructs proto message Bare Metal Standalone Cluster."""
    kwargs = {
        'name': self._standalone_cluster_name(args),
        'description': getattr(args, 'description', None),
        'bareMetalVersion': getattr(args, 'version', None),
        'networkConfig': self._network_config(args),
        'loadBalancer': self._load_balancer_config(args),
        'controlPlane': self._control_plane_config(args),
        'clusterOperations': self._cluster_operations_config(args),
        'maintenanceConfig': self._maintenance_config(args),
        'securityConfig': self._security_config(args),
        'nodeAccessConfig': self._node_access_config(args),
        'annotations': self._annotations(args),
        'binaryAuthorization': self._binary_authorization(args),
    }

    return self._set_config_if_exists(
        messages.BareMetalStandaloneCluster, kwargs
    )

  def _set_config_if_exists(self, config_type, kwargs):
    if any(kwargs.values()):
      return config_type(**kwargs)
    else:
      return None


class StandaloneClustersClient(_BareMetalStandaloneClusterClient):
  """Client for clusters in gkeonprem bare metal standalone API."""

  def __init__(self, **kwargs):
    super(StandaloneClustersClient, self).__init__(**kwargs)
    self._service = self._client.projects_locations_bareMetalStandaloneClusters

  def List(self, location_ref, limit=None, page_size=None):
    """Lists Clusters in the GKE On-Prem Bare Metal Standalone API."""
    list_req = messages.GkeonpremProjectsLocationsBareMetalStandaloneClustersListRequest(
        parent=location_ref.RelativeName()
    )

    return list_pager.YieldFromList(
        self._service,
        list_req,
        field='bareMetalStandaloneClusters',
        batch_size=page_size,
        limit=limit,
        batch_size_attribute='pageSize')

  def Describe(self, resource_ref):
    """Gets a GKE On-Prem Bare Metal Standalone API cluster resource."""
    req = messages.GkeonpremProjectsLocationsBareMetalStandaloneClustersGetRequest(
        name=resource_ref.RelativeName()
    )

    return self._service.Get(req)

  def Enroll(self, args: parser_extensions.Namespace):
    """Enrolls an existing bare metal standalone cluster to the GKE on-prem API within a given project and location."""
    kwargs = {
        'membership': self._standalone_cluster_membership_name(args),
        'bareMetalStandaloneClusterId': self._standalone_cluster_id(args),
    }
    req = messages.GkeonpremProjectsLocationsBareMetalStandaloneClustersEnrollRequest(
        parent=self._standalone_cluster_parent(args),
        enrollBareMetalStandaloneClusterRequest=messages.EnrollBareMetalStandaloneClusterRequest(
            **kwargs
        ),
    )

    return self._service.Enroll(req)

  def Unenroll(self, args: parser_extensions.Namespace):
    """Unenrolls an Anthos on bare metal standalone cluster."""
    kwargs = {
        'name': self._standalone_cluster_name(args),
        'allowMissing': getattr(args, 'allow_missing', None),
        'ignoreErrors': getattr(args, 'ignore_errors', None),
    }
    req = messages.GkeonpremProjectsLocationsBareMetalStandaloneClustersUnenrollRequest(
        **kwargs
    )

    return self._service.Unenroll(req)

  def QueryVersionConfig(self, args: parser_extensions.Namespace):
    """Query Anthos on bare metal standalone cluster version configuration."""
    kwargs = {
        'upgradeConfig_clusterName': self._standalone_cluster_name(args),
        'parent': self._location_ref(args).RelativeName(),
    }

    # This is a workaround for the limitation in apitools with nested messages.
    encoding.AddCustomJsonFieldMapping(
        messages.GkeonpremProjectsLocationsBareMetalStandaloneClustersQueryVersionConfigRequest,
        'upgradeConfig_clusterName',
        'upgradeConfig.clusterName',
    )

    req = messages.GkeonpremProjectsLocationsBareMetalStandaloneClustersQueryVersionConfigRequest(
        **kwargs
    )
    return self._service.QueryVersionConfig(req)

  def Update(self, args: parser_extensions.Namespace):
    """Updates an Anthos on bare metal standalone cluster."""
    kwargs = {
        'name': self._standalone_cluster_name(args),
        'allowMissing': getattr(args, 'allow_missing', None),
        'updateMask': update_mask.get_update_mask(
            args, update_mask.BARE_METAL_STANDALONE_CLUSTER_ARGS_TO_UPDATE_MASKS
        ),
        'validateOnly': getattr(args, 'validate_only', False),
        'bareMetalStandaloneCluster': self._bare_metal_standalone_cluster(args),
    }
    req = messages.GkeonpremProjectsLocationsBareMetalStandaloneClustersPatchRequest(
        **kwargs
    )

    return self._service.Patch(req)



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command group for Network Security Security Profile Groups."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base


@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA)
class SecurityProfileGroups(base.Group):
  """Manage Network Security - Security Profile Groups.

  Manage Network Security - Security Profile Groups.

  ## EXAMPLES

  To create a Security Profile Group with the name `my-security-profile-group`
  (Either a fully specified path or `--location` and `--organization` flags for
  SPG should be specified), `--threat-prevention-profile` `my-security-profile`
  and optional description as `optional description`, run:

      $ {command} create my-security-profile-group --organization=1234
      --location=global
      --threat-prevention-profile=`organizations/1234/locations/global/securityProfiles/my-security-profile`
      --description='optional description'

  To delete an Security Profile Group called `my-security-profile-group` (Either
  a fully specified path or `--location` and `--organization` flags for SPG
  should be specified) run:

      $ {command} delete my-security-profile-group --organization=1234
      --location=global

  To show details of a Security Profile Group named `my-security-profile-group`
  (Either a fully specified path or `--location` and `--organization` flags for
  SPG should be specified) run:

      $ {command} describe my-security-profile-group --organization=1234
      --location=global

  To list Security Profile Groups in specifed location and organization, run:

      $ {command} list --location=global

  To update an SPG with new Threat prevention profile `my-new-security-profile`
  (Either a fully specified path or `--location` and `--organization` flags for
  SPG should be specified), run:

      $ {command} update my-security-profile-group --organization=1234
      --location=global
      --threat-prevention-profile=`organizations/1234/locations/global/securityProfiles/my-new-security-profile`
      --description='New Security Profile of type threat prevention'
  """
  category = base.NETWORK_SECURITY_CATEGORY



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Useful commands for interacting with the Cloud Firestore Backup Schedules API."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.firestore import api_utils
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import exceptions as ex


def _GetBackupSchedulesService():
  """Returns the service to interact with the Firestore Backup Schedules."""
  return api_utils.GetClient().projects_databases_backupSchedules


def GetBackupSchedule(project, database, backup_schedule):
  """Gets a backup schedule.

  Args:
    project: the project of the database of the backup schedule, a string.
    database: the database id of the backup schedule, a string.
    backup_schedule: the backup schedule to read, a string.

  Returns:
    a backup schedule.
  """
  messages = api_utils.GetMessages()
  return _GetBackupSchedulesService().Get(
      messages.FirestoreProjectsDatabasesBackupSchedulesGetRequest(
          name='projects/{}/databases/{}/backupSchedules/{}'.format(
              project,
              database,
              backup_schedule,
          ),
      )
  )


def ListBackupSchedules(project, database):
  """Lists backup schedules under a database.

  Args:
    project: the project of the database of the backup schedule, a string.
    database: the database id of the backup schedule, a string.

  Returns:
    a list of backup schedules.
  """
  messages = api_utils.GetMessages()
  return list(
      _GetBackupSchedulesService()
      .List(
          messages.FirestoreProjectsDatabasesBackupSchedulesListRequest(
              parent='projects/{}/databases/{}'.format(
                  project,
                  database,
              ),
          )
      )
      .backupSchedules
  )


def DeleteBackupSchedule(project, database, backup_schedule):
  """Deletes a backup schedule.

  Args:
    project: the project of the database of the backup schedule, a string.
    database: the database id of the backup schedule, a string.
    backup_schedule: the backup schedule to delete, a string.

  Returns:
    Empty response message.
  """
  messages = api_utils.GetMessages()
  return _GetBackupSchedulesService().Delete(
      messages.FirestoreProjectsDatabasesBackupSchedulesDeleteRequest(
          name='projects/{}/databases/{}/backupSchedules/{}'.format(
              project,
              database,
              backup_schedule,
          ),
      )
  )


def UpdateBackupSchedule(project, database, backup_schedule, retention):
  """Updates a backup schedule.

  Args:
    project: the project of the database of the backup schedule, a string.
    database: the database id of the backup schedule, a string.
    backup_schedule: the backup to read, a string.
    retention: the retention of the backup schedule, an int. At what relative
      time in the future, compared to the creation time of the backup should the
      backup be deleted. The unit is seconds.

  Returns:
    a backup schedule.
  """
  messages = api_utils.GetMessages()
  backup_schedule_updates = messages.GoogleFirestoreAdminV1BackupSchedule()
  if retention:
    backup_schedule_updates.retention = api_utils.FormatDurationString(
        retention
    )

  return _GetBackupSchedulesService().Patch(
      messages.FirestoreProjectsDatabasesBackupSchedulesPatchRequest(
          name='projects/{}/databases/{}/backupSchedules/{}'.format(
              project,
              database,
              backup_schedule,
          ),
          googleFirestoreAdminV1BackupSchedule=backup_schedule_updates,
      )
  )


def CreateBackupSchedule(
    project, database, retention, recurrence, day_of_week=None
):
  """Creates a backup schedule.

  Args:
    project: the project of the database of the backup schedule, a string.
    database: the database id of the backup schedule, a string.
    retention: the retention of the backup schedule, an int. At what relative
      time in the future, compared to the creation time of the backup should the
      backup be deleted. The unit is seconds.
    recurrence: the recurrence of the backup schedule, a string. The valid
      values are: daily and weekly.
    day_of_week: day of week for weekly backup schdeule.

  Returns:
    a backup schedule.

  Raises:
    InvalidArgumentException: if recurrence is invalid.
    ConflictingArgumentsException: if recurrence is daily but day-of-week is
    provided.
    RequiredArgumentException: if recurrence is weekly but day-of-week is not
    provided.
  """
  messages = api_utils.GetMessages()
  backup_schedule = messages.GoogleFirestoreAdminV1BackupSchedule()
  backup_schedule.retention = api_utils.FormatDurationString(retention)
  if recurrence == 'daily':
    if day_of_week is not None:
      raise ex.ConflictingArgumentsException(
          '--day-of-week',
          'Cannot set day of week for daily backup schedules.',
      )
    backup_schedule.dailyRecurrence = (
        messages.GoogleFirestoreAdminV1DailyRecurrence()
    )
  elif recurrence == 'weekly':
    if day_of_week is None:
      raise ex.RequiredArgumentException(
          '--day-of-week',
          'Day of week is required for weekly backup schedules, please use'
          ' --day-of-week to specify this value',
      )
    backup_schedule.weeklyRecurrence = (
        messages.GoogleFirestoreAdminV1WeeklyRecurrence()
    )
    backup_schedule.weeklyRecurrence.day = ConvertDayOfWeek(day_of_week)
  else:
    raise ex.InvalidArgumentException(
        '--recurrence',
        'invalid recurrence: {}. The available values are: `daily` and'
        ' `weekly`.'.format(recurrence),
    )

  return _GetBackupSchedulesService().Create(
      messages.FirestoreProjectsDatabasesBackupSchedulesCreateRequest(
          parent='projects/{}/databases/{}'.format(
              project,
              database,
          ),
          googleFirestoreAdminV1BackupSchedule=backup_schedule,
      )
  )


def ConvertDayOfWeek(day):
  """Converts the user-given day-of-week into DayValueValuesEnum.

  Args:
    day: day of Week for weekly backup schdeule.

  Returns:
    DayValueValuesEnum.

  Raises:
    ValueError: if it is an invalid input.
  """

  day_num = arg_parsers.DayOfWeek.DAYS.index(day)
  messages = api_utils.GetMessages().GoogleFirestoreAdminV1WeeklyRecurrence()
  # a special case for SUN.
  if day_num == 0:
    day_num = 7
  return messages.DayValueValuesEnum(day_num)



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Util methods for Stackdriver Monitoring Surface."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import re

from apitools.base.py import encoding
from googlecloudsdk.calliope import exceptions as calliope_exc
from googlecloudsdk.command_lib.projects import util as projects_util
from googlecloudsdk.command_lib.util.apis import arg_utils
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import resources
from googlecloudsdk.core import yaml
from googlecloudsdk.core.util import times
import six


CHANNELS_FIELD_REMAPPINGS = {'channelLabels': 'labels'}

SNOOZE_FIELD_DELETIONS = ['criteria']

MIGRATED_FROM_PROMETHEUS_TEXT = (
    'Notification channel migrated from Prometheus alert manager file'
)


class YamlOrJsonLoadError(exceptions.Error):
  """Exception for when a JSON or YAML string could not loaded as a message."""


class NoUpdateSpecifiedError(exceptions.Error):
  """Exception for when user passes no arguments that specifies an update."""


class ConditionNotFoundError(exceptions.Error):
  """Indiciates the Condition the user specified does not exist."""


class ConflictingFieldsError(exceptions.Error):
  """Inidicates that the JSON or YAML string have conflicting fields."""


class MonitoredProjectNameError(exceptions.Error):
  """Inidicates that an invalid Monitored Project name has been specified."""


class MissingRequiredFieldError(exceptions.Error):
  """Inidicates that supplied policy/alert rule is missing required field(s)."""


def ValidateUpdateArgsSpecified(args, update_arg_dests, resource):
  if not any([args.IsSpecified(dest) for dest in update_arg_dests]):
    raise NoUpdateSpecifiedError(
        'Did not specify any flags for updating the {}.'.format(resource))


def _RemapFields(yaml_obj, field_remappings):
  for field_name, remapped_name in six.iteritems(field_remappings):
    if field_name in yaml_obj:
      if remapped_name in yaml_obj:
        raise ConflictingFieldsError('Cannot specify both {} and {}.'.format(
            field_name, remapped_name))
      yaml_obj[remapped_name] = yaml_obj.pop(field_name)
  return yaml_obj


def _DeleteFields(yaml_obj, field_deletions):
  for field_name in field_deletions:
    if field_name in yaml_obj:
      yaml_obj.pop(field_name)
  return yaml_obj


def MessageFromString(msg_string, message_type, display_type,
                      field_remappings=None, field_deletions=None):
  try:
    msg_as_yaml = yaml.load(msg_string)
    if field_remappings:
      msg_as_yaml = _RemapFields(msg_as_yaml, field_remappings)
    if field_deletions:
      msg_as_yaml = _DeleteFields(msg_as_yaml, field_deletions)
    msg = encoding.PyValueToMessage(message_type, msg_as_yaml)
    return msg
  except Exception as exc:  # pylint: disable=broad-except
    raise YamlOrJsonLoadError(
        'Could not parse YAML or JSON string for [{0}]: {1}'.format(
            display_type, exc))


def _FlagToDest(flag_name):
  """Converts a --flag-arg to its dest name."""
  return flag_name[len('--'):].replace('-', '_')


def _FormatDuration(duration):
  return '{}s'.format(duration)


def GetBasePolicyMessageFromArgs(args, policy_class):
  """Returns the base policy from args."""
  if args.IsSpecified('policy') or args.IsSpecified('policy_from_file'):
    # Policy and policy_from_file are in a mutex group.
    policy_string = args.policy or args.policy_from_file
    policy = MessageFromString(policy_string, policy_class, 'AlertPolicy')
  else:
    policy = policy_class()
  return policy


def CheckConditionArgs(args):
  """Checks if condition arguments exist and are specified correctly.

  Args:
    args: argparse.Namespace, the parsed arguments.
  Returns:
    bool: True, if '--condition-filter' is specified.
  Raises:
    RequiredArgumentException:
      if '--if' is not set but '--condition-filter' is specified.
    InvalidArgumentException:
      if flag in should_not_be_set is specified without '--condition-filter'.
  """

  if args.IsSpecified('condition_filter'):
    if not args.IsSpecified('if_value'):
      raise calliope_exc.RequiredArgumentException(
          '--if',
          'If --condition-filter is set then --if must be set as well.')
    return True
  else:
    should_not_be_set = [
        '--aggregation',
        '--duration',
        '--trigger-count',
        '--trigger-percent',
        '--condition-display-name',
        '--if',
        '--combiner'
    ]
    for flag in should_not_be_set:
      if flag == '--if':
        dest = 'if_value'
      else:
        dest = _FlagToDest(flag)
      if args.IsSpecified(dest):
        raise calliope_exc.InvalidArgumentException(
            flag,
            'Should only be specified if --condition-filter is also specified.')
    return False


def BuildCondition(messages, condition=None, display_name=None,
                   aggregations=None, trigger_count=None,
                   trigger_percent=None, duration=None, condition_filter=None,
                   if_value=None):
  """Populates the fields of a Condition message from args.

  Args:
    messages: module, module containing message classes for the stackdriver api
    condition: Condition or None, a base condition to populate the fields of.
    display_name: str, the display name for the condition.
    aggregations: list[Aggregation], list of Aggregation messages for the
      condition.
    trigger_count: int, corresponds to the count field of the condition trigger.
    trigger_percent: float, corresponds to the percent field of the condition
      trigger.
    duration: int, The amount of time that a time series must fail to report new
      data to be considered failing.
    condition_filter: str, A filter that identifies which time series should be
      compared with the threshold.
    if_value: tuple[str, float] or None, a tuple containing a string value
      corresponding to the comparison value enum and a float with the condition
      threshold value. None indicates that this should be an Absence condition.

  Returns:
    Condition, a condition with its fields populated from the args
  """
  if not condition:
    condition = messages.Condition()

  if display_name is not None:
    condition.displayName = display_name

  trigger = None
  if trigger_count or trigger_percent:
    trigger = messages.Trigger(
        count=trigger_count, percent=trigger_percent)

  kwargs = {
      'trigger': trigger,
      'duration': duration,
      'filter': condition_filter,
  }

  # This should be unset, not None, if empty
  if aggregations:
    kwargs['aggregations'] = aggregations

  if if_value is not None:
    comparator, threshold_value = if_value  # pylint: disable=unpacking-non-sequence
    if not comparator:
      condition.conditionAbsent = messages.MetricAbsence(**kwargs)
    else:
      comparison_enum = messages.MetricThreshold.ComparisonValueValuesEnum
      condition.conditionThreshold = messages.MetricThreshold(
          comparison=getattr(comparison_enum, comparator),
          thresholdValue=threshold_value,
          **kwargs)

  return condition


def ParseNotificationChannel(channel_name, project=None):
  project = project or properties.VALUES.core.project.Get(required=True)
  return resources.REGISTRY.Parse(
      channel_name, params={'projectsId': project},
      collection='monitoring.projects.notificationChannels')


def ModifyAlertPolicy(base_policy, messages, display_name=None, combiner=None,
                      documentation_content=None, documentation_format=None,
                      enabled=None, channels=None, field_masks=None):
  """Override and/or add fields from other flags to an Alert Policy."""
  if field_masks is None:
    field_masks = []

  if display_name is not None:
    field_masks.append('display_name')
    base_policy.displayName = display_name

  if ((documentation_content is not None or documentation_format is not None)
      and not base_policy.documentation):
    base_policy.documentation = messages.Documentation()
  if documentation_content is not None:
    field_masks.append('documentation.content')
    base_policy.documentation.content = documentation_content
  if documentation_format is not None:
    field_masks.append('documentation.mime_type')
    base_policy.documentation.mimeType = documentation_format

  if enabled is not None:
    field_masks.append('enabled')
    base_policy.enabled = enabled

  # None indicates no update and empty list indicates we want to explicitly set
  # an empty list.
  if channels is not None:
    field_masks.append('notification_channels')
    base_policy.notificationChannels = channels

  if combiner is not None:
    field_masks.append('combiner')
    combiner = arg_utils.ChoiceToEnum(
        combiner, base_policy.CombinerValueValuesEnum, item_type='combiner')
    base_policy.combiner = combiner


def ValidateAtleastOneSpecified(args, flags):
  if not any([args.IsSpecified(_FlagToDest(flag))
              for flag in flags]):
    raise calliope_exc.MinimumArgumentException(flags)


def CreateAlertPolicyFromArgs(args, messages):
  """Builds an AleryPolicy message from args."""
  policy_base_flags = ['--display-name', '--policy', '--policy-from-file']
  ValidateAtleastOneSpecified(args, policy_base_flags)

  # Get a base policy object from the flags
  policy = GetBasePolicyMessageFromArgs(args, messages.AlertPolicy)
  combiner = args.combiner if args.IsSpecified('combiner') else None
  enabled = args.enabled if args.IsSpecified('enabled') else None
  channel_refs = args.CONCEPTS.notification_channels.Parse() or []
  channels = [channel.RelativeName() for channel in channel_refs] or None
  documentation_content = args.documentation or args.documentation_from_file
  documentation_format = (
      args.documentation_format if documentation_content else None)
  ModifyAlertPolicy(
      policy,
      messages,
      display_name=args.display_name,
      combiner=combiner,
      documentation_content=documentation_content,
      documentation_format=documentation_format,
      enabled=enabled,
      channels=channels)

  if CheckConditionArgs(args):
    aggregations = None
    if args.aggregation:
      aggregations = [MessageFromString(
          args.aggregation, messages.Aggregation, 'Aggregation')]

    condition = BuildCondition(
        messages,
        display_name=args.condition_display_name,
        aggregations=aggregations,
        trigger_count=args.trigger_count,
        trigger_percent=args.trigger_percent,
        duration=_FormatDuration(args.duration),
        condition_filter=args.condition_filter,
        if_value=args.if_value)
    policy.conditions.append(condition)

  return policy


def GetConditionFromArgs(args, messages):
  """Builds a Condition message from args."""
  condition_base_flags = ['--condition-filter', '--condition',
                          '--condition-from-file']
  ValidateAtleastOneSpecified(args, condition_base_flags)

  condition = None
  condition_string = args.condition or args.condition_from_file
  if condition_string:
    condition = MessageFromString(
        condition_string, messages.Condition, 'Condition')

  aggregations = None
  if args.aggregation:
    aggregations = [MessageFromString(
        args.aggregation, messages.Aggregation, 'Aggregation')]

  return BuildCondition(
      messages,
      condition=condition,
      display_name=args.condition_display_name,
      aggregations=aggregations,
      trigger_count=args.trigger_count,
      trigger_percent=args.trigger_percent,
      duration=_FormatDuration(args.duration),
      condition_filter=args.condition_filter,
      if_value=args.if_value)


def GetConditionFromPolicy(condition_name, policy):
  for condition in policy.conditions:
    if condition.name == condition_name:
      return condition

  raise ConditionNotFoundError(
      'No condition with name [{}] found in policy.'.format(condition_name))


def RemoveConditionFromPolicy(condition_name, policy):
  for i, condition in enumerate(policy.conditions):
    if condition.name == condition_name:
      policy.conditions.pop(i)
      return policy

  raise ConditionNotFoundError(
      'No condition with name [{}] found in policy.'.format(condition_name))


def ModifyNotificationChannel(base_channel, channel_type=None, enabled=None,
                              display_name=None, description=None,
                              field_masks=None):
  """Modifies base_channel's properties using the passed arguments."""
  if field_masks is None:
    field_masks = []

  if channel_type is not None:
    field_masks.append('type')
    base_channel.type = channel_type
  if display_name is not None:
    field_masks.append('display_name')
    base_channel.displayName = display_name
  if description is not None:
    field_masks.append('description')
    base_channel.description = description
  if enabled is not None:
    field_masks.append('enabled')
    base_channel.enabled = enabled
  return base_channel


def GetNotificationChannelFromArgs(args, messages):
  """Builds a NotificationChannel message from args."""
  channels_base_flags = ['--display-name', '--channel-content',
                         '--channel-content-from-file']
  ValidateAtleastOneSpecified(args, channels_base_flags)

  channel_string = args.channel_content or args.channel_content_from_file
  if channel_string:
    channel = MessageFromString(channel_string, messages.NotificationChannel,
                                'NotificationChannel',
                                field_remappings=CHANNELS_FIELD_REMAPPINGS)
    # Without this, labels will be in a random order every time.
    if channel.labels:
      channel.labels.additionalProperties = sorted(
          channel.labels.additionalProperties, key=lambda prop: prop.key)
  else:
    channel = messages.NotificationChannel()

  enabled = args.enabled if args.IsSpecified('enabled') else None
  return ModifyNotificationChannel(channel,
                                   channel_type=args.type,
                                   display_name=args.display_name,
                                   description=args.description,
                                   enabled=enabled)


def ParseCreateLabels(labels, labels_cls):
  return encoding.DictToAdditionalPropertyMessage(
      labels, labels_cls, sort_items=True)


def ProcessUpdateLabels(args, labels_name, labels_cls, orig_labels):
  """Returns the result of applying the diff constructed from args.

  This API doesn't conform to the standard patch semantics, and instead does
  a replace operation on update. Therefore, if there are no updates to do,
  then the original labels must be returned as writing None into the labels
  field would replace it.

  Args:
    args: argparse.Namespace, the parsed arguments with update_labels,
      remove_labels, and clear_labels
    labels_name: str, the name for the labels flag.
    labels_cls: type, the LabelsValue class for the new labels.
    orig_labels: message, the original LabelsValue value to be updated.

  Returns:
    LabelsValue: The updated labels of type labels_cls.

  Raises:
    ValueError: if the update does not change the labels.
  """
  labels_diff = labels_util.Diff(
      additions=getattr(args, 'update_' + labels_name),
      subtractions=getattr(args, 'remove_' + labels_name),
      clear=getattr(args, 'clear_' + labels_name))
  if not labels_diff.MayHaveUpdates():
    return None

  return labels_diff.Apply(labels_cls, orig_labels).GetOrNone()


def ParseMonitoredProject(monitored_project_name, project_fallback):
  """Returns the metrics scope and monitored project.

  Parse the specified monitored project name and return the metrics scope and
  monitored project.

  Args:
    monitored_project_name: The name of the monitored project to create/delete.
    project_fallback: When set, allows monitored_project_name to be just a
      project id or number.

  Raises:
    MonitoredProjectNameError: If an invalid monitored project name is
    specified.

  Returns:
     (metrics_scope_def, monitored_project_def): Project parsed metrics scope
       project id, Project parsed metrics scope project id
  """
  matched = re.match(
      'locations/global/metricsScopes/([a-z0-9:\\-]+)/projects/([a-z0-9:\\-]+)',
      monitored_project_name)
  if matched:
    if matched.group(0) != monitored_project_name:
      raise MonitoredProjectNameError(
          'Invalid monitored project name has been specified.')
    # full name
    metrics_scope_def = projects_util.ParseProject(matched.group(1))
    monitored_project_def = projects_util.ParseProject(matched.group(2))
  else:
    metrics_scope_def = projects_util.ParseProject(
        properties.VALUES.core.project.Get(required=True))
    monitored_resource_container_matched = re.match(
        'projects/([a-z0-9:\\-]+)', monitored_project_name
    )
    if monitored_resource_container_matched:
      monitored_project_def = projects_util.ParseProject(
          monitored_resource_container_matched.group(1)
      )
    elif project_fallback:
      log.warning(
          'Received an incorrectly formatted project name. Expected '
          '"projects/{identifier}" received "{identifier}". Assuming '
          'given resource is a project.'.format(
              identifier=monitored_project_name
          )
      )
      monitored_project_def = projects_util.ParseProject(monitored_project_name)
    else:
      raise MonitoredProjectNameError(
          'Invalid monitored project name has been specified.'
      )
  return metrics_scope_def, monitored_project_def


def ParseMonitoredResourceContainer(
    monitored_resource_container_name, project_fallback
):
  """Returns the monitored resource container identifier.

  Parse the specified monitored_resource_container_name and return the
  identifier.

  Args:
    monitored_resource_container_name: The monitored resource container. Ex -
      projects/12345.
    project_fallback: When set, allows monitored_resource_container_name to be
      just a project id or number.

  Raises:
    MonitoredProjectNameError: If an invalid monitored project name is
    specified.

  Returns:
     resource_type, monitored_resource_container_identifier: Monitored resource
     container type and identifier
  """
  matched = re.match(
      '(projects)/([a-z0-9:\\-]+)', monitored_resource_container_name
  )
  if matched:
    return matched.group(1), matched.group(2)
  elif project_fallback:
    log.warning(
        'Received an incorrectly formatted project name. Expected '
        '"projects/{identifier}" received "{identifier}". Assuming '
        'given resource is a project.'.format(
            identifier=monitored_resource_container_name
        )
    )
    return (
        'projects',
        projects_util.ParseProject(monitored_resource_container_name).Name(),
    )
  else:
    raise MonitoredProjectNameError(
        'Invalid monitored project name has been specified.'
    )


def ParseSnooze(snooze_name, project=None):
  project = project or properties.VALUES.core.project.Get(required=True)
  return resources.REGISTRY.Parse(
      snooze_name,
      params={'projectsId': project},
      collection='monitoring.projects.snoozes',
  )


def GetBaseSnoozeMessageFromArgs(args, snooze_class, update=False):
  """Returns the base snooze from args."""
  if args.IsSpecified('snooze_from_file'):
    snooze_string = args.snooze_from_file
    if update:
      snooze = MessageFromString(
          snooze_string,
          snooze_class,
          'Snooze',
          field_deletions=SNOOZE_FIELD_DELETIONS,
      )
    else:
      snooze = MessageFromString(
          snooze_string,
          snooze_class,
          'Snooze',
      )
  else:
    snooze = snooze_class()
  return snooze


def ModifySnooze(
    base_snooze,
    messages,
    display_name=None,
    criteria_policies=None,
    start_time=None,
    end_time=None,
    field_masks=None,
):
  """Override and/or add fields from other flags to an Snooze."""
  if field_masks is None:
    field_masks = []

  start_time_target = None
  start_time_from_base = False
  if start_time is not None:
    field_masks.append('interval.start_time')
    start_time_target = start_time
  else:
    try:
      start_time_target = times.ParseDateTime(base_snooze.interval.startTime)
      start_time_from_base = True
    except AttributeError:
      pass

  end_time_target = None
  end_time_from_base = False
  if end_time is not None:
    field_masks.append('interval.end_time')
    end_time_target = end_time
  else:
    try:
      end_time_target = times.ParseDateTime(base_snooze.interval.endTime)
      end_time_from_base = True
    except AttributeError:
      pass

  try:
    if start_time_target is not None and not start_time_from_base:
      base_snooze.interval.startTime = times.FormatDateTime(start_time_target)
    if end_time_target is not None and not end_time_from_base:
      base_snooze.interval.endTime = times.FormatDateTime(end_time_target)
  except AttributeError:
    interval = messages.TimeInterval()
    interval.startTime = times.FormatDateTime(start_time_target)
    interval.endTime = times.FormatDateTime(end_time_target)
    base_snooze.interval = interval

  if display_name is not None:
    field_masks.append('display_name')
    base_snooze.displayName = display_name

  if criteria_policies is not None:
    field_masks.append('criteria_policies')
    criteria = messages.Criteria()
    criteria.policies = criteria_policies
    base_snooze.criteria = criteria


def CreateSnoozeFromArgs(args, messages):
  """Builds a Snooze message from args."""
  snooze_base_flags = ['--display-name', '--snooze-from-file']
  ValidateAtleastOneSpecified(args, snooze_base_flags)

  # Get a base snooze object from the flags
  snooze = GetBaseSnoozeMessageFromArgs(args, messages.Snooze)

  ModifySnooze(
      snooze,
      messages,
      display_name=args.display_name,
      criteria_policies=args.criteria_policies,
      start_time=args.start_time,
      end_time=args.end_time)

  return snooze


# Conversions from interval suffixes to number of seconds.
# (m => 60s, d => 86400s, etc)
_INTERVAL_CONV_DICT = {'s': 1}
_INTERVAL_CONV_DICT['ms'] = 0.001 * _INTERVAL_CONV_DICT['s']
_INTERVAL_CONV_DICT['m'] = 60 * _INTERVAL_CONV_DICT['s']
_INTERVAL_CONV_DICT['h'] = 60 * _INTERVAL_CONV_DICT['m']
_INTERVAL_CONV_DICT['d'] = 24 * _INTERVAL_CONV_DICT['h']
_INTERVAL_CONV_DICT['w'] = 7 * _INTERVAL_CONV_DICT['d']
_INTERVAL_CONV_DICT['y'] = 365 * _INTERVAL_CONV_DICT['d']
_INTERVAL_PART_REGEXP = re.compile(
    '^([0-9]+)(%s)' % '|'.join(_INTERVAL_CONV_DICT)
)


def ConvertIntervalToSeconds(interval):
  """Forked from datelib.py.

  Convert a formatted Prometheus string representing an interval into seconds.

  The accepted interval string syntax is:
    interval: (interval_part)*
    interval_part: decimal_integer unit
    unit: "ms"       # Milliseconds
        | "s"        # Seconds
        | "m"        # Minutes
        | "h"        # Hours
        | "d"        # Days
        | "w"        # Weeks (7 days)
        | "y"        # Years (365 days)

  No whitespace is allowed.
  The empty string is valid (and equivalent to 0 seconds).
  |decimal_integer| cannot include a sign.
  No endianness ordering is required when using multiple interval_part-s. For
  example, "1s1Y" and "1Y1s" are both valid.

  Examples:
    "45m" = 45 minutes
    "14d12h" = 14 days + 12 hours
    "5d12h30m" = 5 days + 12 hours + 30 minutes

  Args:
    interval: String to interpret as an interval.  See above for a description
      of the syntax.

  Raises:
    ValueError: If the provided time_string contains unexpected
    characters.

  Returns:
    A non-negative integer representing the number of seconds represented by the
    interval string.
  """
  total = 0
  original_interval = interval
  # The initial value of "previous_multiplier" is larger than any valid
  # multiplier.
  previous_multiplier = _INTERVAL_CONV_DICT.get('y') + 1
  while interval:
    # Match the interval_part at the prefix of "interval".
    match = _INTERVAL_PART_REGEXP.match(interval)
    if not match:
      raise ValueError(
          '{} is invalid due to missing unit of time or unexpected'
          ' character(s).'.format(original_interval)
      )

    try:
      num = int(match.group(1))
    except ValueError:
      raise ValueError(
          'Found invalid character in {}, which is neither an integer nor unit'
          ' of time.'.format(original_interval)
      )

    # The time unit suffix should always exist. Otherwise, the previous match()
    # would have failed.
    suffix = match.group(2)
    multiplier = _INTERVAL_CONV_DICT.get(suffix)
    if multiplier >= previous_multiplier:
      # Time units must be ordered from largest to smallest.
      raise ValueError(
          'Time units not ordered from largest to smallest in {}.'.format(
              original_interval
          )
      )

    previous_multiplier = multiplier
    num *= multiplier

    total += num
    # Remove the interval_part prefix from "interval".
    interval = interval[match.end(0) :]
  return total


def ConvertPrometheusTimeStringToEvaluationDurationInSeconds(time_string):
  """Converts Prometheus time to duration JSON string.

  Args:
    time_string: String provided by the alert rule YAML file defining time
      (ex:1h30m)

  Raises:
    ValueError: If the provided time_string is not a multiple of 30 seconds or
    is less than 30 seconds.

  Returns:
    Duration proto string representing the adjusted seconds (multiple of 30
    seconds) value of the provided time_string
  """
  seconds = ConvertIntervalToSeconds(time_string)
  if seconds < 30:
    raise ValueError(
        '{time_string} converted to {seconds}s is less than 30 seconds.'.format(
            time_string=time_string,
            seconds=seconds,
        )
    )
  elif seconds % 30 != 0:
    raise ValueError(
        '{} converted to {}s is not a multiple of 30 seconds.'.format(
            time_string, seconds,
        )
    )
  return _FormatDuration(seconds)


# Regular expressions for matching common Prometheus templating language
# constructs.
_VALUE_VARIABLE_REGEXP = re.compile(r'\{\{ *(humanize )? *\$value *\}\}')
_LABELS_VARIABLE_REGEXP = re.compile(r'\{\{ *(humanize )? *\$labels *\}\}')
_LABELS_KEY_REGEXP = re.compile(
    r'\{\{ *(humanize )? *\$labels\.([a-zA-Z_][a-zA-Z0-9_]*) *\}\}')


def TranslatePromQLTemplateToDocumentVariables(template):
  """Translate Prometheus templating language constructs to document variables.

  TranslatePromQLTemplateToDocumentVariables translates common Prometheus
  templating language constructs to their equivalent Cloud Alerting document
  variables. See:
  https://prometheus.io/docs/prometheus/latest/configuration/template_reference/
  and https://cloud.google.com/monitoring/alerts/doc-variables.

  Only the following constructs will be translated:

  "{{ $value }}" will be translated to "${metric.label.value}".
  "{{ humanize $value }}" will be translated to "${metric.label.value}".
  "{{ $labels.<name> }}" will be translated to
  "${metric_or_resource.label.<name>}".
  "{{ humanize $labels.<name> }}" will be translated to
  "${metric_or_resource.label.<name>}".
  "{{ $labels }}" will be translated to
  "${metric_or_resource.labels}".
  "{{ humanize $labels }}" will be translated to
  "${metric_or_resource.labels}".

  The number of spaces inside the curly braces is immaterial.

  All other Prometheus templating language constructs are not translated.

  Notes:
  1. A document variable reference that does not match a variable
     will be rendered as "(none)".
  2. We do not know whether a {{ $labels.<name> }} construct refers to
     a Cloud Alerting metric or a resource label. Thus we translate it to
     "${metric_or_resource.label.<name>}".
     Note that a reference to a non-existent label will be rendered as "(none)".

  Examples:
  1. "[{{$labels.a}}] VALUE = {{ $value }}" will be translated to
     "[${metric_or_resource.label.a}] VALUE = ${metric.label.value}".

  2. "[{{humanize $labels.a}}] VALUE = {{ humanize $value }}"
     will be translated to
     "[${metric_or_resource.label.a}] VALUE = ${metric.label.value}".

  Args:
    template: String contents of the "subject" or "content" fields of an
    AlertPolicy protoco buffer. The contents of these fields is a template
    which may contain Prometheus templating language constructs.

  Returns:
    The translated template.
  """
  return _LABELS_KEY_REGEXP.sub(
      r'${metric_or_resource.label.\2}',
      _LABELS_VARIABLE_REGEXP.sub(
          r'${metric_or_resource.labels}',
          _VALUE_VARIABLE_REGEXP.sub(
              r'${metric.label.value}', template)))


# A regular expression matching a valid Prometheus label name.
# See: https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels
_VALID_LABEL_REGEXP = re.compile(r'[A-Za-z_][A-Za-z0-9_]*')


def BuildPrometheusCondition(messages, group, rule):
  """Populates Alert Policy conditions translated from a Prometheus alert rule.

  Args:
    messages: Object containing information about all message types allowed.
    group: Information about the parent group of the current rule.
    rule: The current alert rule being translated into an Alert Policy.

  Raises:
    MissingRequiredFieldError: If the provided group/rule is missing an required
    field needed for translation.
    ValueError: If the provided rule name is not a valid Prometheus label name.

  Returns:
     The Alert Policy condition corresponding to the Prometheus group and rule
     provided.
  """
  condition = messages.Condition()
  condition.conditionPrometheusQueryLanguage = (
      messages.PrometheusQueryLanguageCondition()
  )
  if group.get('name') is None:
    raise MissingRequiredFieldError(
        'Missing rule group name in field group.name'
    )
  if rule.get('alert') is None:
    raise MissingRequiredFieldError(
        'Missing alert rule name in field group.rules.alert'
    )
  if _VALID_LABEL_REGEXP.fullmatch(rule.get('alert')) is None:
    raise ValueError(
        'An invalid alert rule name in field group.rules.alert '
        '(not a valid PromQL label name)'
    )
  if rule.get('expr') is None:
    raise MissingRequiredFieldError(
        'Missing a PromQL expression in field groups.rules.expr'
    )
  condition.conditionPrometheusQueryLanguage.ruleGroup = group.get('name')
  condition.displayName = rule.get('alert')
  condition.conditionPrometheusQueryLanguage.alertRule = rule.get('alert')
  condition.conditionPrometheusQueryLanguage.query = rule.get('expr')

  # optional fields
  if rule.get('for') is not None:
    condition.conditionPrometheusQueryLanguage.duration = _FormatDuration(
        ConvertIntervalToSeconds(rule.get('for'))
    )
  if group.get('interval') is not None:
    condition.conditionPrometheusQueryLanguage.evaluationInterval = (
        ConvertPrometheusTimeStringToEvaluationDurationInSeconds(
            group.get('interval')
        )
    )
  if rule.get('labels') is not None:
    condition.conditionPrometheusQueryLanguage.labels = (
        messages.PrometheusQueryLanguageCondition.LabelsValue()
    )
    for k, v in rule.get('labels').items():
      condition.conditionPrometheusQueryLanguage.labels.additionalProperties.append(
          messages.PrometheusQueryLanguageCondition.LabelsValue.AdditionalProperty(
              key=k, value=v
          )
      )

  return condition


def PrometheusMessageFromString(rule_yaml, messages, channels):
  """Populates Alert Policies translated from Prometheus alert rules.

  Args:
    rule_yaml: Opened object of the Prometheus YAML file provided.
    messages: Object containing information about all message types allowed.
    channels: List of Notification Channel names to be added to the translated
      policies.

  Raises:
    YamlOrJsonLoadError: If the YAML file cannot be loaded.

  Returns:
     A list of the Alert Policies corresponding to the Prometheus rules YAML
     file provided.
  """
  try:
    contents = yaml.load(rule_yaml)
    if contents is None:
      raise ValueError('Failed to load YAML file. Is it empty?')

    policies = []
    if contents.get('groups') is None:
      raise ValueError('No groups')

    for group in contents.get('groups'):
      if group.get('rules') is None:
        raise ValueError('No rules in group "%s"' % group.get('name'))

      for rule in group.get('rules'):
        condition = BuildPrometheusCondition(messages, group, rule)
        policy = messages.AlertPolicy()
        policy.conditions.append(condition)
        if rule.get('annotations') is not None:
          policy.documentation = messages.Documentation()
          if rule.get('annotations').get('subject') is not None:
            policy.documentation.subject = (
                TranslatePromQLTemplateToDocumentVariables(
                    rule.get('annotations').get('subject')))
          if rule.get('annotations').get('description') is not None:
            policy.documentation.content = (
                TranslatePromQLTemplateToDocumentVariables(
                    rule.get('annotations').get('description')))
          policy.documentation.mimeType = 'text/markdown'

        if _VALID_LABEL_REGEXP.fullmatch(group.get('name')) is not None:
          # The rule group name is a valid Prometheus label name.
          policy.displayName = '{0}/{1}'.format(
              group.get('name'), rule.get('alert')
          )
        else:
          # The rule group name is NOT a valid Prometheus label name.
          policy.displayName = '"{0}"/{1}'.format(
              group.get('name'), rule.get('alert')
          )

        policy.combiner = arg_utils.ChoiceToEnum(
            'OR', policy.CombinerValueValuesEnum, item_type='combiner'
        )

        if channels is not None:
          policy.notificationChannels = channels

        policies.append(policy)

    return policies
  except Exception as exc:  # pylint: disable=broad-except
    raise YamlOrJsonLoadError('Could not parse YAML: {0}'.format(exc))


def CreateBasePromQLNotificationChannel(channel_name, messages):
  """Helper function for creating a basic Notification Channel translated from a Prometheus alert_manager YAML.

  Args:
    channel_name: The display name of the desired channel.
    messages: Object containing information about all message types allowed.

  Returns:
     A base Notification Channel containing the requested display_name and
     other basic fields.
  """
  channel = messages.NotificationChannel()
  channel.displayName = channel_name
  channel.description = MIGRATED_FROM_PROMETHEUS_TEXT
  channel.labels = messages.NotificationChannel.LabelsValue()
  return channel


def BuildChannelsFromPrometheusReceivers(receiver_config, messages):
  """Populates a Notification Channel translated from Prometheus alert manager.

  Args:
    receiver_config: Object containing information the Prometheus receiver. For
      example receiver_configs, see
      https://github.com/prometheus/alertmanager/blob/main/doc/examples/simple.yml
    messages: Object containing information about all message types allowed.

  Raises:
    MissingRequiredFieldError: If the provided alert manager file contains
    receivers with missing required field(s).

  Returns:
     The Notification Channel corresponding to the Prometheus alert manager
     provided.
  """
  channels = []
  channel_name = receiver_config.get('name')
  if channel_name is None:
    raise MissingRequiredFieldError(
        'Supplied alert manager file contains receiver without a required field'
        ' "name"'
    )

  if receiver_config.get('email_configs') is not None:
    for fields in receiver_config.get('email_configs'):
      if fields.get('to') is not None:
        channel = CreateBasePromQLNotificationChannel(channel_name, messages)
        channel.type = 'email'
        channel.labels.additionalProperties.append(
            messages.NotificationChannel.LabelsValue.AdditionalProperty(
                key='email_address', value=fields.get('to')
            )
        )
        channels.append(channel)

  if receiver_config.get('pagerduty_configs') is not None:
    for fields in receiver_config.get('pagerduty_configs'):
      if fields.get('service_key') is not None:
        channel = CreateBasePromQLNotificationChannel(channel_name, messages)
        channel.type = 'pagerduty'
        channel.labels.additionalProperties.append(
            messages.NotificationChannel.LabelsValue.AdditionalProperty(
                key='service_key', value=fields.get('service_key')
            )
        )
        channels.append(channel)

  if receiver_config.get('webhook_configs') is not None:
    for fields in receiver_config.get('webhook_configs'):
      if fields.get('url') is not None:
        channel = CreateBasePromQLNotificationChannel(channel_name, messages)
        channel.type = 'webhook_tokenauth'
        channel.labels.additionalProperties.append(
            messages.NotificationChannel.LabelsValue.AdditionalProperty(
                key='url', value=fields.get('url')
            )
        )
        channels.append(channel)

  # Tell users that their defined receiver type is not supported by the
  # migration tool and will not be translated.
  # TODO(b/277099361): Have a continue prompt telling users that certain
  # receiver types are not supported.
  supported_receiver_fields = set(
      ['name', 'email_configs', 'pagerduty_configs', 'webhook_configs']
  )
  for field in receiver_config.keys():
    if field not in supported_receiver_fields:
      log.out.Print(
          'Found unsupported receiver type {field}. {name}.{field} will not be'
          ' translated.'.format(field=field, name=receiver_config.get('name'))
      )

  return channels


def NotificationChannelMessageFromString(alert_manager_yaml, messages):
  """Populates Alert Policies translated from Prometheus alert rules.

  Args:
    alert_manager_yaml: Opened object of the Prometheus YAML file provided.
    messages: Object containing information about all message types allowed.

  Raises:
    YamlOrJsonLoadError: If the YAML file cannot be loaded.

  Returns:
     The Alert Policies corresponding to the Prometheus rules YAML file
     provided.
  """
  try:
    contents = yaml.load(alert_manager_yaml)
  except Exception as exc:  # pylint: disable=broad-except
    raise YamlOrJsonLoadError('Could not parse YAML: {0}'.format(exc))

  channels = []
  for receiver_config in contents.get('receivers'):
    channels += BuildChannelsFromPrometheusReceivers(receiver_config, messages)
  return channels


def CreatePromQLPoliciesFromArgs(args, messages, channels=None):
  """Builds a PromQL policies message from args.

  Args:
    args: Flags provided by the user.
    messages: Object containing information about all message types allowed.
    channels: List of full Notification Channel names ("projects/<>/...") to be
      added to the translated policies.

  Returns:
     The Alert Policies corresponding to the Prometheus rules YAML file
     provided. In the case that no file is specified, the default behavior is to
     return an empty list.
  """

  if args.IsSpecified('policies_from_prometheus_alert_rules_yaml'):
    all_rule_yamls = args.policies_from_prometheus_alert_rules_yaml
    policies = []
    for rule_yaml in all_rule_yamls:
      policies += PrometheusMessageFromString(rule_yaml, messages, channels)
  else:
    policies = []

  return policies


def CreateNotificationChannelsFromArgs(args, messages):
  """Builds a notification channel message from args.

  Args:
    args: Flags provided by the user.
    messages: Object containing information about all message types allowed.

  Returns:
     The notification channels corresponding to the Prometheus alert manager
     YAML file provided. In the case that no file is specified, the default
     behavior is to return an empty list.
  """

  if args.IsSpecified('channels_from_prometheus_alertmanager_yaml'):
    alert_manager_yaml = args.channels_from_prometheus_alertmanager_yaml
    channels = NotificationChannelMessageFromString(
        alert_manager_yaml, messages
    )
  else:
    channels = []
  return channels



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags for Firewall Plus Endpoint commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.network_security.firewall_endpoints import activation_api
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs


ENDPOINT_RESOURCE_NAME = "FIREWALL_ENDPOINT"


def AddEndpointResource(release_track, parser):
  """Adds Firewall Plus endpoint resource."""
  api_version = activation_api.GetApiVersion(release_track)
  resource_spec = concepts.ResourceSpec(
      "networksecurity.organizations.locations.firewallEndpoints",
      "firewall endpoint",
      api_version=api_version,
      organizationsId=concepts.ResourceParameterAttributeConfig(
          "organization",
          "Organization ID of the {resource}.",
          parameter_name="organizationsId",
      ),
      locationsId=concepts.ResourceParameterAttributeConfig(
          "zone",
          "Zone of the {resource}.",
          parameter_name="locationsId",
      ),
      firewallEndpointsId=concepts.ResourceParameterAttributeConfig(
          "endpoint-name",
          "Name of the {resource}",
          parameter_name="firewallEndpointsId",
      ),
  )
  presentation_spec = presentation_specs.ResourcePresentationSpec(
      name=ENDPOINT_RESOURCE_NAME,
      concept_spec=resource_spec,
      required=True,
      group_help="Firewall Plus.",
  )
  return concept_parsers.ConceptParser([presentation_spec]).AddToParser(parser)


def AddMaxWait(
    parser,
    default_max_wait,
    help_text="Time to synchronously wait for the operation to complete, after which the operation continues asynchronously. Ignored if --no-async isn't specified. See $ gcloud topic datetimes for information on time formats.",
):
  """Adds --max-wait flag."""
  parser.add_argument(
      "--max-wait",
      dest="max_wait",
      required=False,
      default=default_max_wait,
      help=help_text,
      type=arg_parsers.Duration(),
  )


def MakeGetUriFunc(release_track):
  return (
      lambda x: activation_api.GetEffectiveApiEndpoint(release_track) + x.name
  )


def AddOrganizationArg(parser, help_text="Organization of the endpoint"):
  parser.add_argument("--organization", required=True, help=help_text)


def AddDescriptionArg(parser, help_text="Description of the endpoint"):
  parser.add_argument("--description", required=False, help=help_text)


def AddZoneArg(parser, required=True, help_text="Zone of the endpoint"):
  parser.add_argument("--zone", required=required, help=help_text)



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manages content bundles for Policy Controller."""
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base as calliope_base
from googlecloudsdk.command_lib.container.fleet.features import base
from googlecloudsdk.command_lib.container.fleet.policycontroller import command
from googlecloudsdk.command_lib.container.fleet.policycontroller import content


@calliope_base.Hidden
@calliope_base.ReleaseTracks(
    calliope_base.ReleaseTrack.ALPHA, calliope_base.ReleaseTrack.BETA
)
class Set(base.UpdateCommand, command.PocoCommand):
  """Sets bundle installation for Policy Controller content.

  Google-defined policy bundles of constraints can be installed onto Policy
  Controller installations.

  The namespace exclusion flag (`--exempted-namespaces`) will specify a set of
  namespaces that the installed bundle will ignore. Subsequent calls with the
  same bundle name and this flag will overwrite what namespaces are being
  ignored. Using `--no-exempted-namespaces` or specifying no namespaces with
  `--exempted-namespaces` will remove all namespaces from the ignore list.

  To uninstall a bundle, use the `remove` command.

  ## Examples

  To install a policy bundle:

    $ {command} cis-k8s-v1.5.1

  To install a policy bundle, while ignoring (exempting) certain namespaces from
  being affected by the bundle:

    $ {command} cis-k8s-v1.5.1
        --exempted-namespaces=kube-system,gatekeeper-system

  To remove all exempted namespaces from a particular bundles ignore list:

    $ {command} cis-k8s-v1.5.1 --no-exempted-namespaces
  """

  feature_name = 'policycontroller'

  @classmethod
  def Args(cls, parser):
    cmd_flags = content.Flags(parser, 'bundles')
    cmd_flags.add_memberships()
    cmd_flags.add_exempted_namespaces()

    parser.add_argument(
        content.ARG_LABEL_BUNDLE,
        help='The constraint bundle to install in Policy Controller.',
    )

  def Run(self, args):
    parser = content.FlagParser(args, self.messages)
    specs = self.path_specs(args, True)
    updated_specs = {path: self.modify(s, parser) for path, s in specs.items()}
    return self.update_specs(updated_specs)

  def modify(self, spec, parser):
    policy_content = (
        spec.policycontroller.policyControllerHubConfig.policyContent
    )
    spec.policycontroller.policyControllerHubConfig.policyContent = (
        parser.install_bundle(policy_content)
    )
    return spec



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Type aliases for Cloud Functions v2 API."""
from typing import Union

from googlecloudsdk.generated_clients.apis.cloudfunctions.v2 import cloudfunctions_v2_messages as v2_messages
from googlecloudsdk.generated_clients.apis.cloudfunctions.v2alpha import cloudfunctions_v2alpha_messages as v2alpha_messages
from googlecloudsdk.generated_clients.apis.cloudfunctions.v2beta import cloudfunctions_v2beta_messages as v2beta_messages


BuildConfig = Union[
    v2_messages.BuildConfig,
    v2alpha_messages.BuildConfig,
    v2beta_messages.BuildConfig,
]
EventTrigger = Union[
    v2_messages.EventTrigger,
    v2alpha_messages.EventTrigger,
    v2beta_messages.EventTrigger,
]
ServiceConfig = Union[
    v2_messages.ServiceConfig,
    v2alpha_messages.ServiceConfig,
    v2beta_messages.ServiceConfig,
]
Source = Union[
    v2_messages.Source, v2alpha_messages.Source, v2beta_messages.Source
]
Function = Union[
    v2_messages.Function, v2alpha_messages.Function, v2beta_messages.Function
]


# Enum types (these unfortunately can't be resolved from the type aliases above)
IngressSettings = Union[
    v2_messages.ServiceConfig.IngressSettingsValueValuesEnum,
    v2beta_messages.ServiceConfig.IngressSettingsValueValuesEnum,
    v2alpha_messages.ServiceConfig.IngressSettingsValueValuesEnum,
]
LabelsValue = Union[
    v2_messages.Function.LabelsValue,
    v2alpha_messages.Function.LabelsValue,
    v2beta_messages.Function.LabelsValue,
]
RetryPolicy = Union[
    v2_messages.EventTrigger.RetryPolicyValueValuesEnum,
    v2alpha_messages.EventTrigger.RetryPolicyValueValuesEnum,
    v2beta_messages.EventTrigger.RetryPolicyValueValuesEnum,
]
VpcConnectorEgressSettings = Union[
    v2_messages.ServiceConfig.VpcConnectorEgressSettingsValueValuesEnum,
    v2alpha_messages.ServiceConfig.VpcConnectorEgressSettingsValueValuesEnum,
    v2beta_messages.ServiceConfig.VpcConnectorEgressSettingsValueValuesEnum,
]



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to SSH into a Cloud TPU VM Node."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import argparse
import os.path

from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute import completers
from googlecloudsdk.command_lib.compute import flags
from googlecloudsdk.command_lib.compute import ssh_utils
from googlecloudsdk.command_lib.compute.tpus.tpu_vm import ssh as tpu_ssh_utils
from googlecloudsdk.core import properties


def AddCommandArgGroup(parser):
  """Argument group for running commands using SSH."""
  command_group = parser.add_argument_group(
      help='These arguments are used to run commands using SSH.')
  command_group.add_argument(
      '--command',
      help="""\
      Command to run on the Cloud TPU VM.

      Runs the command on the target Cloud TPU VM and then exits.

      Note: in the case of a TPU Pod, it will only run the command in the
      workers specified with the `--worker` flag (defaults to worker 0 if not
      set).
      """)
  command_group.add_argument(
      '--output-directory',
      help="""\
      Path to the directory to output the logs of the commands.

      The path can be relative or absolute. The directory must already exist.

      If not specified, standard output will be used.

      The logs will be written in files named {WORKER_ID}.log. For example:
      "2.log".
      """)


def AddSSHArgs(parser):
  """Additional flags and positional args to be passed to *ssh(1)*."""
  parser.add_argument(
      '--ssh-flag',
      action='append',
      help="""\
      Additional flags to be passed to *ssh(1)*. It is recommended that flags
      be passed using an assignment operator and quotes. Example:

        $ {command} example-instance --zone=us-central1-a --ssh-flag="-vvv" --ssh-flag="-L 80:localhost:80"

      This flag will replace occurences of ``%USER%'' and ``%TPU%'' with
      their dereferenced values. For example, passing ``80:%TPU%:80`` into
      the flag is equivalent to passing ``80:162.222.181.197:80'' to *ssh(1)*
      if the external IP address of 'example-instance' is 162.222.181.197.

      If connecting to the instance's external IP, then %TPU% is replaced
      with that, otherwise it is replaced with the internal IP.
      """)

  parser.add_argument(
      'user_tpu',
      completer=completers.InstancesCompleter,
      metavar='[USER@]TPU',
      help="""\
      Specifies the Cloud TPU VM to SSH into.

      ``USER'' specifies the username with which to SSH. If omitted, the user
      login name is used.

      ``TPU'' specifies the name of the Cloud TPU VM to SSH into.
      """)

  parser.add_argument(
      'ssh_args',
      nargs=argparse.REMAINDER,
      help="""\
          Flags and positionals passed to the underlying ssh implementation.
          """,
      example="""\
        $ {command} example-instance --zone=us-central1-a -- -vvv -L 80:%TPU%:80
      """)


@base.ReleaseTracks(base.ReleaseTrack.GA)
class Ssh(base.Command):
  """SSH into a Cloud TPU VM."""

  # IAP and Batching are not available for GA.
  _ENABLE_IAP = False
  _ENABLE_BATCHING = False

  @classmethod
  def Args(cls, parser):
    """Set up arguments for this command.

    Args:
      parser: An argparse.ArgumentParser.
    """
    ssh_utils.BaseSSHCLIHelper.Args(parser)
    AddSSHArgs(parser)
    tpu_ssh_utils.AddTPUSSHArgs(parser, enable_iap=cls._ENABLE_IAP,
                                enable_batching=cls._ENABLE_BATCHING)
    AddCommandArgGroup(parser)
    flags.AddZoneFlag(parser, resource_type='tpu', operation_type='ssh')

  def Run(self, args):
    user, tpu_name = ssh_utils.GetUserAndInstance(args.user_tpu)

    # If zone is not set, retrieve the one from the config.
    if args.zone is None:
      args.zone = properties.VALUES.compute.zone.Get(required=True)
    # Validate the output path.
    if args.output_directory:
      if not args.command:
        raise exceptions.InvalidArgumentException(
            '--output_directory', 'cannot be specified without the `--command` '
            'flag. Please specify the `--command` flag or remove the '
            '--output-directory flag.')
      output_directory_path = os.path.abspath(
          os.path.expandvars(os.path.expanduser(args.output_directory)))
      if not os.path.isdir(output_directory_path):
        raise exceptions.InvalidArgumentException(
            '--output_directory', 'Failed to find directory {}. Please create '
            'it or specify another directory'.format(output_directory_path))

    username_requested = '@' in args.user_tpu
    prepped_node = [None]
    tpu_ssh_utils.PrepareNodeForSSH(
        tpu_name,
        user,
        args,
        self.ReleaseTrack(),
        self._ENABLE_BATCHING,
        username_requested,
        prepped_node,
        0,
    )

    ssh_batch_size = 1
    if self._ENABLE_BATCHING and prepped_node[0]:
      ssh_batch_size = tpu_ssh_utils.ParseBatchSize(
          args.batch_size, len(prepped_node[0].worker_ips)
      )

    tpu_ssh_utils.SSHIntoPreppedNodes(
        prepped_node,
        args,
        ssh_batch_size,
    )


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class SshAlpha(Ssh):
  """SSH into a Cloud TPU VM (Alpha)."""
  _ENABLE_IAP = True
  _ENABLE_BATCHING = True


Ssh.detailed_help = {
    'brief':
        'SSH into a Cloud TPU VM.',
    'EXAMPLES':
        """
        To SSH into a Cloud TPU VM, run:

            $ {command} my-tpu

        To SSH into worker 1 on a Cloud TPU VM Pod, run:

            $ {command} my-tpu --worker=1

        To run an SSH command in a Cloud TPU VM (for example, to print the
        time since last boot), run:

            $ {command} my-tpu --command="last boot"

        To run the same command in all workers in a Cloud TPU VM simultaneously,
        run:

            $ {command} my-tpu --command="last boot" --worker=all
        """
}



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for the cloud deploy custom target type resource."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.clouddeploy import custom_target_type


def PatchCustomTargetType(resource):
  """Patches a custom target type resource.

  Args:
    resource: apitools.base.protorpclite.messages.Message, custom target type
      message.

  Returns:
    The operation message
  """
  return custom_target_type.CustomTargetTypesClient().Patch(resource)


def DeleteCustomTargetType(name):
  """Deletes a custom target type resource.

  Args:
    name: str, custom target type name.

  Returns:
    The operation message
  """
  return custom_target_type.CustomTargetTypesClient().Delete(name)



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Marketplace Solution instance delete command."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.mps.mps_client import MpsClient
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.mps import flags
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import resources

DETAILED_HELP = {
    'DESCRIPTION':
        """
          Delete a Marketplace Solution instance.
        """,
    'EXAMPLES':
        """
          To delete an instance called ``my-instance'' in region
          ``us-central1'', run:

          $ {command} my-instance  --region=us-central1
        """,
}


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Delete(base.DeleteCommand):
  """Delete a Marketplace Solution instance."""

  @staticmethod
  def Args(parser):
    """Register flags for this command."""
    flags.AddInstanceArgToParser(parser, positional=True)

  def Run(self, args):
    instance = args.CONCEPTS.instance.Parse()
    client = MpsClient()
    product = properties.VALUES.mps.product.Get(required=True)

    op_ref = client.DeleteInstance(product, instance)
    if op_ref.done:
      log.DeletedResource(instance.Name(), kind='Instance')
      return op_ref

    op_resource = resources.REGISTRY.ParseRelativeName(
        op_ref.name,
        collection='marketplacesolutions.projects.locations.operations',
        api_version='v1alpha1')
    poller = waiter.CloudOperationPollerNoResources(client.operation_service)
    res = waiter.WaitFor(
        poller, op_resource,
        'Waiting for operation [{}] to complete'.format(op_ref.name))
    log.DeletedResource(instance.Name(), kind='Instance')
    return res


Delete.detailed_help = DETAILED_HELP



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Task for retrieving a list of resources from the cloud.

Typically executed in a task iterator:
googlecloudsdk.command_lib.storage.tasks.task_executor.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from googlecloudsdk.command_lib.storage import list_util
from googlecloudsdk.command_lib.storage.list_util import BaseFormatWrapper
from googlecloudsdk.command_lib.storage.list_util import BaseListExecutor
from googlecloudsdk.command_lib.storage.list_util import DisplayDetail
from googlecloudsdk.command_lib.storage.resources import gcloud_full_resource_formatter
from googlecloudsdk.command_lib.storage.resources import gsutil_full_resource_formatter
from googlecloudsdk.command_lib.storage.resources import resource_reference
from googlecloudsdk.command_lib.storage.resources import resource_util
from googlecloudsdk.command_lib.storage.resources import shim_format_util


LONG_LIST_ROW_FORMAT = (
    '{size:>10}  {creation_time:>20}  {url}{metageneration}{etag}'
)


class _HeaderFormatWrapper(BaseFormatWrapper):
  """For formatting how headers print when listed."""

  def __init__(
      self,
      resource,
      all_versions=False,
      display_detail=DisplayDetail.SHORT,
      include_etag=False,
      readable_sizes=False,
      full_formatter=None,
      use_gsutil_style=False,
  ):
    """See BaseFormatWrapper class for function doc strings."""

    super(_HeaderFormatWrapper, self).__init__(
        resource,
        all_versions=all_versions,
        display_detail=display_detail,
        full_formatter=full_formatter,
        include_etag=include_etag,
        readable_sizes=readable_sizes,
        use_gsutil_style=use_gsutil_style,
    )

  def __str__(self):
    if self._use_gsutil_style and isinstance(
        self.resource, resource_reference.BucketResource
    ):
      # Gsutil does not show header lines for buckets.
      return ''

    url = self.resource.storage_url.versionless_url_string
    if self._display_detail == DisplayDetail.JSON:
      return self.resource.get_json_dump()
    # This will print as "gs://bucket:" or "gs://bucket/prefix/:".
    return '\n{}:'.format(url)


class _ResourceFormatWrapper(BaseFormatWrapper):
  """For formatting how resources print when listed."""

  def __init__(
      self,
      resource,
      all_versions=False,
      display_detail=DisplayDetail.SHORT,
      include_etag=False,
      readable_sizes=False,
      full_formatter=None,
      use_gsutil_style=False,
  ):
    """See BaseFormatWrapper class for function doc strings."""

    super(_ResourceFormatWrapper, self).__init__(
        resource,
        all_versions=all_versions,
        display_detail=display_detail,
        include_etag=include_etag,
        readable_sizes=readable_sizes,
        use_gsutil_style=use_gsutil_style,
    )

    self._full_formatter = full_formatter

  def _format_for_list_long(self):
    """Returns string of select properties from resource."""
    if isinstance(self.resource, resource_reference.PrefixResource):
      # Align PrefixResource URLs with ObjectResource URLs.
      return LONG_LIST_ROW_FORMAT.format(
          size='',
          creation_time='',
          url=self.resource.storage_url.url_string,
          metageneration='',
          etag='',
      )

    creation_time = resource_util.get_formatted_timestamp_in_utc(
        self.resource.creation_time
    )

    url_string, metageneration_string = self._check_and_handles_all_versions()

    if self._include_etag:
      etag_string = '  etag={}'.format(str(self.resource.etag))
    else:
      etag_string = ''

    # Full example (add 9 spaces of padding to the left):
    # 8  2020-07-27T20:58:25Z  gs://b/o  metageneration=4  etag=CJqt6aup7uoCEAQ=
    return LONG_LIST_ROW_FORMAT.format(
        size=list_util.check_and_convert_to_readable_sizes(
            self.resource.size, self._readable_sizes, self._use_gsutil_style
        ),
        creation_time=creation_time,
        url=url_string,
        metageneration=metageneration_string,
        etag=etag_string,
    )

  def __str__(self):
    if self._display_detail == DisplayDetail.LONG and (
        isinstance(self.resource, resource_reference.ObjectResource)
        or isinstance(self.resource, resource_reference.PrefixResource)
    ):
      return self._format_for_list_long()
    if self._display_detail == DisplayDetail.FULL and (
        isinstance(self.resource, resource_reference.BucketResource)
        or isinstance(self.resource, resource_reference.ObjectResource)
    ):
      return self._full_formatter.format(
          self.resource, show_version_in_url=self._all_versions
      )
    if self._display_detail == DisplayDetail.JSON:
      return self.resource.get_json_dump()
    if self._all_versions:
      # Include generation in URL.
      return self.resource.storage_url.url_string
    return self.resource.storage_url.versionless_url_string


class LsExecutor(BaseListExecutor):
  """Helper class for the ls command."""

  def __init__(
      self,
      cloud_urls,
      all_versions=False,
      buckets_flag=False,
      display_detail=DisplayDetail.SHORT,
      fetch_encrypted_object_hashes=False,
      halt_on_empty_response=True,
      include_etag=False,
      next_page_token=None,
      readable_sizes=False,
      recursion_flag=False,
      soft_deleted_only=False,
      use_gsutil_style=False,
  ):
    """See BaseListExecutor class for function doc strings."""
    super(LsExecutor, self).__init__(
        cloud_urls=cloud_urls,
        all_versions=all_versions,
        buckets_flag=buckets_flag,
        display_detail=display_detail,
        fetch_encrypted_object_hashes=fetch_encrypted_object_hashes,
        halt_on_empty_response=halt_on_empty_response,
        include_etag=include_etag,
        next_page_token=next_page_token,
        readable_sizes=readable_sizes,
        recursion_flag=recursion_flag,
        soft_deleted_only=soft_deleted_only,
        use_gsutil_style=use_gsutil_style,
    )

    if use_gsutil_style:
      self._full_formatter = (
          gsutil_full_resource_formatter.GsutilFullResourceFormatter()
      )
    else:
      self._full_formatter = (
          gcloud_full_resource_formatter.GcloudFullResourceFormatter()
      )
    self._header_wrapper = _HeaderFormatWrapper
    self._object_wrapper = _ResourceFormatWrapper

  def _print_summary_for_top_level_url(
      self, resource_url, only_display_buckets, object_count, total_bytes
  ):
    if (
        self._display_detail in (DisplayDetail.LONG, DisplayDetail.FULL)
        and not only_display_buckets
    ):
      # Long listing needs summary line.
      print(
          'TOTAL: {} objects, {} bytes ({})'.format(
              object_count,
              int(total_bytes),
              shim_format_util.get_human_readable_byte_value(
                  total_bytes, self._use_gsutil_style
              ),
          )
      )

  def _print_json_list(self, resource_wrappers):
    """Prints ResourceWrapper objects as JSON list."""
    is_empty_list = True
    for i, resource_wrapper in enumerate(resource_wrappers):
      is_empty_list = False
      if i == 0:
        # Start of JSON list for long long listing.
        print('[')
        print(resource_wrapper, end='')
      else:
        # Print resource without newline at end to allow list formatting for
        # unknown number of items in generator.
        print(',\n{}'.format(resource_wrapper), end='')

    # New line because we were removing it from previous prints to give us
    # the ability to do a trailing comma for JSON list printing.
    print()
    if not is_empty_list:
      # Close long long listing JSON list. Prints nothing if no items.
      print(']')



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create a new version."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.cloudkms import base as cloudkms_base
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.kms import exceptions as kms_exceptions
from googlecloudsdk.command_lib.kms import flags
from googlecloudsdk.core import log


@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA,
                    base.ReleaseTrack.GA)
class Create(base.CreateCommand):
  r"""Create a new version.

  Creates a new version within the given key.

  ## EXAMPLES

  The following command creates a new version within the `frodo`
  key, `fellowship` keyring, and `global` location and sets it as
  the primary version:

    $ {command} --location=global \
        --keyring=fellowship \
        --key=frodo --primary

  The following command creates a new version within the `legolas`
  key, `fellowship` keyring, `us-central1` location,
  `https://example.kms/v0/some/key/path` as the address for its external key,
  and sets it as the key's primary version:

    $ {command} --location=us-central1 \
        --keyring=fellowship \
        --key=legolas \
        --external-key-uri=https://example.kms/v0/some/key/path \
        --primary

  The following command creates a new version within the `bilbo`
  key, `fellowship` keyring, `us-central1` location,
  `v0/some/key/path` as the ekm connection key path for its external key,
  and sets it as the key's primary version:

    $ {command} --location=us-central1 \
        --keyring=fellowship \
        --key=bilbo \
        --ekm-connection-key-path=v0/some/key/path \
        --primary
  """

  GOOGLE_SYMMETRIC_ENCRYPTION = (
      cloudkms_base.GetMessagesModule().CryptoKeyVersion.AlgorithmValueValuesEnum.GOOGLE_SYMMETRIC_ENCRYPTION
  )

  SYMMETRIC_NEW_PRIMARY_MESSAGE = (
      'Successfully created key version [{version}] and set it as the primary '
      'version. Future encryption requests will use [{version}] until the next '
      'key rotation. Data that was encrypted with an older key version can '
      'still be decrypted, and Cloud KMS will automatically choose the correct '
      'key for decryption based on the ciphertext.')

  @staticmethod
  def Args(parser):
    flags.AddKeyResourceFlags(parser)
    flags.AddExternalKeyUriFlag(parser)
    flags.AddEkmConnectionKeyPathFlag(parser)
    parser.add_argument(
        '--primary',
        action='store_true',
        help=(
            'If specified, immediately makes the new version primary. '
            'This should only be used with the `encryption` purpose.'
        ),
    )

  def _CreateCreateCKVRequest(self, args):
    # pylint: disable=line-too-long
    messages = cloudkms_base.GetMessagesModule()
    crypto_key_ref = flags.ParseCryptoKeyName(args)

    if args.external_key_uri and args.ekm_connection_key_path:
      raise kms_exceptions.ArgumentError(
          'Can not specify both --external-key-uri and '
          '--ekm-connection-key-path.')

    if args.external_key_uri or args.ekm_connection_key_path:
      return messages.CloudkmsProjectsLocationsKeyRingsCryptoKeysCryptoKeyVersionsCreateRequest(
          parent=crypto_key_ref.RelativeName(),
          cryptoKeyVersion=messages.CryptoKeyVersion(
              externalProtectionLevelOptions=messages
              .ExternalProtectionLevelOptions(
                  externalKeyUri=args.external_key_uri,
                  ekmConnectionKeyPath=args.ekm_connection_key_path)))

    return messages.CloudkmsProjectsLocationsKeyRingsCryptoKeysCryptoKeyVersionsCreateRequest(
        parent=crypto_key_ref.RelativeName())

  def Run(self, args):
    # pylint: disable=line-too-long
    client = cloudkms_base.GetClientInstance()
    ckv = client.projects_locations_keyRings_cryptoKeys_cryptoKeyVersions
    new_ckv = ckv.Create(self._CreateCreateCKVRequest(args))

    if args.primary:
      version_id = new_ckv.name.split('/')[-1]
      crypto_key_ref = flags.ParseCryptoKeyName(args)
      messages = cloudkms_base.GetMessagesModule()

      req = messages.CloudkmsProjectsLocationsKeyRingsCryptoKeysUpdatePrimaryVersionRequest(
          name=crypto_key_ref.RelativeName(),
          updateCryptoKeyPrimaryVersionRequest=(
              messages.UpdateCryptoKeyPrimaryVersionRequest(
                  cryptoKeyVersionId=version_id)))
      client.projects_locations_keyRings_cryptoKeys.UpdatePrimaryVersion(req)

      if new_ckv.algorithm == self.GOOGLE_SYMMETRIC_ENCRYPTION:
        log.err.Print(
            self.SYMMETRIC_NEW_PRIMARY_MESSAGE.format(version=version_id))

    return new_ckv



# -*- coding: utf-8 -*- #
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Delete a secret."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.secrets import api as secrets_api
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.secrets import args as secrets_args
from googlecloudsdk.command_lib.secrets import log as secrets_log
from googlecloudsdk.core.console import console_io


@base.ReleaseTracks(base.ReleaseTrack.GA)
class Delete(base.DeleteCommand):
  r"""Delete a secret.

  Delete a secret and destroy all secret versions. This action is irreversible.
  If the given secret does not exist, this command will succeed, but the
  operation will be a no-op.

  ## EXAMPLES

  Delete a secret 'my-secret':

    $ {command} my-secret

  Delete a secret 'my-secret' using an etag:

    $ {command} my-secret --etag=\"123\"
  """

  CONFIRM_DELETE_MESSAGE = (
      'You are about to destroy the secret [{secret}] and its [{num_versions}] '
      'version(s). This action cannot be reversed.')

  @staticmethod
  def Args(parser):
    secrets_args.AddSecret(
        parser, purpose='to delete', positional=True, required=True)
    secrets_args.AddSecretEtag(parser)

  def Run(self, args):
    messages = secrets_api.GetMessages()
    secret_ref = args.CONCEPTS.secret.Parse()

    # List all secret versions and parse their refs
    versions = secrets_api.Versions().ListWithPager(
        secret_ref=secret_ref, limit=9999)
    active_version_count = 0
    for version in versions:
      if version.state != messages.SecretVersion.StateValueValuesEnum.DESTROYED:
        active_version_count += 1

    msg = self.CONFIRM_DELETE_MESSAGE.format(
        secret=secret_ref.Name(), num_versions=active_version_count)
    console_io.PromptContinue(msg, throw_if_unattended=True, cancel_on_no=True)

    result = secrets_api.Secrets().Delete(secret_ref, etag=args.etag)
    secrets_log.Secrets().Deleted(secret_ref)
    return result


@base.ReleaseTracks(base.ReleaseTrack.BETA)
class DeleteBeta(Delete):
  r"""Delete a secret.

  Delete a secret and destroy all secret versions. This action is irreversible.
  If the given secret does not exist, this command will succeed, but the
  operation will be a no-op.

  ## EXAMPLES

  Delete a secret 'my-secret':

    $ {command} my-secret

  Delete a secret 'my-secret' using etag:

    $ {command} my-secret --etag=\"123\"
  """

  @staticmethod
  def Args(parser):
    secrets_args.AddSecret(
        parser, purpose='to delete', positional=True, required=True)
    secrets_args.AddSecretEtag(parser)

  def Run(self, args):
    messages = secrets_api.GetMessages()
    secret_ref = args.CONCEPTS.secret.Parse()

    # List all secret versions and parse their refs
    versions = secrets_api.Versions().ListWithPager(
        secret_ref=secret_ref, limit=9999)
    active_version_count = 0
    for version in versions:
      if version.state != messages.SecretVersion.StateValueValuesEnum.DESTROYED:
        active_version_count += 1

    msg = self.CONFIRM_DELETE_MESSAGE.format(
        secret=secret_ref.Name(), num_versions=active_version_count)
    console_io.PromptContinue(msg, throw_if_unattended=True, cancel_on_no=True)

    result = secrets_api.Secrets().Delete(secret_ref, etag=args.etag)
    secrets_log.Secrets().Deleted(secret_ref)
    return result



# -*- coding: utf-8 -*- #
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for describing instant snapshot."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute.instant_snapshots import flags as ips_flags


def _CommonArgs(parser):
  DescribeBeta.ips_arg = ips_flags.MakeInstantSnapshotArg(plural=False)
  DescribeBeta.ips_arg.AddArgument(parser, operation_type='describe')


@base.ReleaseTracks(base.ReleaseTrack.BETA)
class DescribeBeta(base.DescribeCommand):
  """Describe a Compute Engine instant snapshot in beta."""

  @classmethod
  def Args(cls, parser):
    _CommonArgs(parser)

  def _Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    client = holder.client
    messages = client.messages

    ips_ref = DescribeBeta.ips_arg.ResolveAsResource(args, holder.resources)

    if ips_ref.Collection() == 'compute.instantSnapshots':
      service = client.apitools_client.instantSnapshots
      request_type = messages.ComputeInstantSnapshotsGetRequest
    elif ips_ref.Collection() == 'compute.regionInstantSnapshots':
      service = client.apitools_client.regionInstantSnapshots
      request_type = messages.ComputeRegionInstantSnapshotsGetRequest

    return client.MakeRequests([(service, 'Get',
                                 request_type(**ips_ref.AsDict()))])

  def Run(self, args):
    return self._Run(args)


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class DescribeAlpha(DescribeBeta):
  """Describe a Compute Engine instant snapshot in alpha."""

  @classmethod
  def Args(cls, parser):
    _CommonArgs(parser)

  def Run(self, args):
    return self._Run(args)


DescribeBeta.detailed_help = {
    'brief': 'Describe a Compute Engine instant snapshot',
    'DESCRIPTION': """\
        *{command}* displays all data associated with a Compute
        Engine instant snapshot in a project.
        """,
    'EXAMPLES': """\
        To describe the instant snapshot 'instant-snapshot-1' in zone 'us-east1-a', run:

            $ {command} instant-snapshot-1 --zone=us-east1-a
        """,
}



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Delete a locally deployed Google Cloud Function."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.functions.local import flags as local_flags
from googlecloudsdk.command_lib.functions.local import util
from googlecloudsdk.core import log

_DETAILED_HELP = {
    'DESCRIPTION': """
        `{command}` Delete a locally deployed Google Cloud Function.
        """,
}


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Delete(base.Command):
  """Delete a locally deployed Google Cloud Function."""

  detailed_help = _DETAILED_HELP

  @staticmethod
  def Args(parser):
    local_flags.AddDeploymentNameFlag(parser)

  def Run(self, args):
    util.ValidateDependencies()
    name = args.NAME[0]
    if not util.ContainerExists(name):
      raise util.ContainerNotFoundException(
          'The container ' + name + ' does not exist.')
    util.RemoveDockerContainer(name)
    log.status.Print('The container ' + name + ' has been deleted.')



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for creating/parsing update resource argument groups."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.calliope.concepts import util as format_util
from googlecloudsdk.command_lib.util.apis import arg_utils
from googlecloudsdk.command_lib.util.apis import update_args
from googlecloudsdk.command_lib.util.apis import yaml_command_schema_util as util
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.core import resources


# TODO(b/280653078) The UX is still under review. These utilities are
# liable to change and should not be used in new surface yet.

# TODO(b/283949482): Place this file in util/args and replace the duplicate
# logic in the util files.


def _GetRelativeNameField(arg_data):
  """Gets message field where the resource's relative name is mapped."""
  api_fields = [
      key
      for key, value in arg_data.resource_method_params.items()
      if util.REL_NAME_FORMAT_KEY in value
  ]
  if not api_fields:
    return None

  return api_fields[0]


def _GetAllSharedAttributes(arg_data, shared_resource_args):
  """Gets a list of all shared resource attributes."""
  if not shared_resource_args:
    ignored_attributes = set()
  else:
    ignored_attributes = set(shared_resource_args)

  # iterate through all attributes except the anchor
  for a in arg_data.attribute_names[:-1]:
    if a in arg_data.removed_flags or concepts.IGNORED_FIELDS.get(a):
      continue
    ignored_attributes.add(a)

  return list(ignored_attributes)


def _GetResourceArgGenerator(arg_data, method, shared_resource_args):
  """Gets a function to generate a resource arg."""
  ignored_attributes = _GetAllSharedAttributes(arg_data, shared_resource_args)
  def ArgGen(name, group_help):
    group_help += '\n\n'
    if arg_data.group_help:
      group_help += arg_data.group_help

    return arg_data.GenerateResourceArg(
        method, name, ignored_attributes, group_help=group_help)
  return ArgGen


class UpdateResourceArgumentGenerator(update_args.UpdateArgumentGenerator):
  """Update flag generator for resource args."""

  @classmethod
  def FromArgData(
      cls, arg_data, method, shared_resource_args=None
  ):
    if arg_data.repeated:
      gen_cls = UpdateListResourceArgumentGenerator
    else:
      gen_cls = UpdateDefaultResourceArgumentGenerator

    arg_name = arg_data.GetAnchorArgName(method)
    is_primary = arg_data.IsPrimaryResource(
        method and method.resource_argument_collection
    )
    if is_primary:
      raise util.InvalidSchemaError(
          '{} is a primary resource. Primary resources are required and '
          'cannot be listed as clearable.'.format(arg_name)
      )

    api_field = _GetRelativeNameField(arg_data)
    if not api_field:
      raise util.InvalidSchemaError(
          '{} does not specify the message field where the relative name is '
          'mapped in resource_method_params. Message field name is needed '
          'in order add update args. Please update '
          'resource_method_params.'.format(arg_name)
      )

    return gen_cls(
        arg_name=arg_name,
        arg_gen=_GetResourceArgGenerator(
            arg_data, method, shared_resource_args),
        api_field=api_field,
        repeated=arg_data.repeated,
        collection=arg_data.collection,
        is_primary=is_primary,
        attribute_flags=arg_utils.GetAttributeFlags(
            arg_data, arg_name, method, shared_resource_args),
    )

  def __init__(
      self,
      arg_name,
      arg_gen=None,
      is_hidden=False,
      api_field=None,
      repeated=False,
      collection=None,
      is_primary=None,
      attribute_flags=None,
  ):
    super(UpdateResourceArgumentGenerator, self).__init__()
    self.arg_name = format_util.NormalizeFormat(arg_name)
    self.arg_gen = arg_gen
    self.is_hidden = is_hidden
    self.api_field = api_field
    self.repeated = repeated
    self.collection = collection
    self.is_primary = is_primary
    self.attribute_flags = attribute_flags

  def _CreateResourceFlag(self, flag_prefix=None, group_help=None):
    flag_name = arg_utils.GetFlagName(
        self.arg_name, flag_prefix=flag_prefix and flag_prefix.value)
    return self.arg_gen(flag_name, group_help=group_help)

  def _RelativeName(self, value):
    return resources.REGISTRY.ParseRelativeName(
        value, self.collection.full_name)

  def GetArgFromNamespace(self, namespace, arg):
    """Retrieves namespace value associated with flag.

    Args:
      namespace: The parsed command line argument namespace.
      arg: base.Argument|concept_parsers.ConceptParser|None, used to get
        namespace value

    Returns:
      value parsed from namespace
    """
    if isinstance(arg, base.Argument):
      return arg_utils.GetFromNamespace(namespace, arg.name)

    if isinstance(arg, concept_parsers.ConceptParser):
      all_anchors = list(arg.specs.keys())
      if len(all_anchors) != 1:
        raise ValueError(
            'ConceptParser must contain exactly one spec for clearable '
            'but found specs {}. {} cannot parse the namespace value if more '
            'than or less than one spec is added to the '
            'ConceptParser.'.format(all_anchors, type(self).__name__))
      name = all_anchors[0]
      value = arg_utils.GetFromNamespace(namespace.CONCEPTS, name)
      if value:
        value = value.Parse()
      return value

    return None

  def GetFieldValueFromMessage(self, existing_message):
    value = arg_utils.GetFieldValueFromMessage(existing_message, self.api_field)
    if not value:
      return None

    if isinstance(value, list):
      return [self._RelativeName(v) for v in value]
    return self._RelativeName(value)

  def Generate(self):
    return super(UpdateResourceArgumentGenerator, self).Generate(
        self.attribute_flags)


class UpdateDefaultResourceArgumentGenerator(UpdateResourceArgumentGenerator):
  """Update flag generator for resource args."""

  @property
  def _empty_value(self):
    return None

  @property
  def set_arg(self):
    return self._CreateResourceFlag(
        group_help='Set {} to new value.'.format(self.arg_name))

  @property
  def clear_arg(self):
    return self._CreateFlag(
        self.arg_name,
        flag_prefix=update_args.Prefix.CLEAR,
        action='store_true',
        help_text='Clear {} value and set to {}.'.format(
            self.arg_name, self._GetTextFormatOfEmptyValue(self._empty_value)),
    )

  def ApplySetFlag(self, output, set_val):
    if set_val:
      return set_val
    return output

  def ApplyClearFlag(self, output, clear_flag):
    if clear_flag:
      return self._empty_value
    return output


class UpdateListResourceArgumentGenerator(UpdateResourceArgumentGenerator):
  """Update flag generator for list resource args."""

  @property
  def _empty_value(self):
    return []

  @property
  def set_arg(self):
    return self._CreateResourceFlag(
        group_help='Set {} to new value.'.format(self.arg_name))

  @property
  def clear_arg(self):
    return self._CreateFlag(
        self.arg_name,
        flag_prefix=update_args.Prefix.CLEAR,
        action='store_true',
        help_text='Clear {} value and set to {}.'.format(
            self.arg_name, self._GetTextFormatOfEmptyValue(self._empty_value)),
    )

  @property
  def update_arg(self):
    return self._CreateResourceFlag(
        flag_prefix=update_args.Prefix.ADD,
        group_help='Add new value to {} list.'.format(self.arg_name))

  @property
  def remove_arg(self):
    return self._CreateResourceFlag(
        flag_prefix=update_args.Prefix.REMOVE,
        group_help='Remove value from {} list.'.format(self.arg_name))

  def ApplySetFlag(self, output, set_val):
    if set_val:
      return set_val
    return output

  def ApplyClearFlag(self, output, clear_flag):
    if clear_flag:
      return self._empty_value
    return output

  def ApplyRemoveFlag(self, existing_val, remove_val):
    value = existing_val or self._empty_value
    if remove_val:
      return [x for x in value if x not in remove_val]
    else:
      return value

  def ApplyUpdateFlag(self, existing_val, update_val):
    value = existing_val or self._empty_value
    if update_val:
      return existing_val + [x for x in update_val if x not in value]
    else:
      return value




# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create a Cloud NetApp Backup Vault."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.netapp.backup_vaults import client as backupvaults_client
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.netapp.backup_vaults import flags as backupvaults_flags
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.core import log


@base.Hidden
@base.ReleaseTracks(base.ReleaseTrack.BETA)
class CreateBeta(base.CreateCommand):
  """Create a Cloud NetApp Backup Vault."""

  _RELEASE_TRACK = base.ReleaseTrack.BETA

  detailed_help = {
      'DESCRIPTION': """\
          Create a Cloud NetApp Backup Vault.
          """,
      'EXAMPLES': """\
          The following command creates a Backup Vault named BACKUP_VAULT asynchronously using the specified arguments:

              $ {command} BACKUP_VAULT --location=LOCATION --async --description="test"
          """,
  }

  @staticmethod
  def Args(parser):
    backupvaults_flags.AddBackupVaultCreateArgs(parser)

  def Run(self, args):
    """Create a Cloud NetApp Backup Vault in the current project."""
    backupvault_ref = args.CONCEPTS.backup_vault.Parse()
    client = backupvaults_client.BackupVaultsClient(self._RELEASE_TRACK)
    labels = labels_util.ParseCreateArgs(
        args, client.messages.BackupVault.LabelsValue
    )
    backup_vault = client.ParseBackupVault(
        name=backupvault_ref.RelativeName(),
        description=args.description,
        labels=labels,
    )
    result = client.CreateBackupVault(
        backupvault_ref, args.async_, backup_vault
    )
    if args.async_:
      command = 'gcloud {} netapp backup-vaults list'.format(
          self.ReleaseTrack().prefix
      )
      log.status.Print(
          'Check the status of the new backup vault by listing all backup'
          ' vaults:\n  $ {} '.format(command)
      )
    return result



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags for Fault Injection Cloud SDK."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.calliope.concepts import deps
from googlecloudsdk.command_lib.fault_injection import location_util
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs


def GetFaultResourceSpec(arg_name='fault', help_text=None):
  """Constructs and returns the Resource specification for Fault."""

  def FaultAttributeConfig():
    return concepts.ResourceParameterAttributeConfig(
        name=arg_name, help_text=help_text
    )

  return concepts.ResourceSpec(
      'faultinjectiontesting.projects.locations.faults',
      resource_name='fault',
      faultsId=FaultAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      locationsId=LocationAttributeConfig(),
  )


def GetFaultResourceArg(
    arg_name='fault', help_text=None, positional=True, required=True
):
  """Constructs and returns the Fault Resource Argument."""

  help_text = help_text or 'Name for the Fault'

  return concept_parsers.ConceptParser.ForResource(
      '{}{}'.format('' if positional else '--', arg_name),
      GetFaultResourceSpec(),
      help_text,
      required=required,
  )


def GetExperimentResourceSpec(arg_name='experiment', help_text=None):
  """Constructs and returns the Resource specification for Fault."""

  def ExperimentAttributeConfig():
    return concepts.ResourceParameterAttributeConfig(
        name=arg_name, help_text=help_text
    )

  return concepts.ResourceSpec(
      'faultinjectiontesting.projects.locations.experiments',
      resource_name='experiment',
      experimentsId=ExperimentAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      locationsId=LocationAttributeConfig(),
  )


def GetExperimentResourceArg(
    arg_name='experiment', help_text=None, positional=True, required=True
):
  """Constructs and returns the Experiment Resource Argument."""

  help_text = help_text or 'Name for the Experiment'

  return concept_parsers.ConceptParser.ForResource(
      '{}{}'.format('' if positional else '--', arg_name),
      GetExperimentResourceSpec(),
      help_text,
      required=required,
  )


def GetJobResourceSpec(arg_name='job', help_text=None):
  """Constructs and returns the Resource specification for Job."""

  def JobAttributeConfig():
    return concepts.ResourceParameterAttributeConfig(
        name=arg_name, help_text=help_text
    )

  return concepts.ResourceSpec(
      'faultinjectiontesting.projects.locations.jobs',
      resource_name='job',
      jobsId=JobAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      locationsId=LocationAttributeConfig(),
  )


def GetJobResourceArg(
    arg_name='job', help_text=None, positional=True, required=True
):
  """Constructs and returns the Job Resource Argument."""

  help_text = help_text or 'Name of the Job'

  return concept_parsers.ConceptParser.ForResource(
      '{}{}'.format('' if positional else '--', arg_name),
      GetJobResourceSpec(),
      help_text,
      required=required,
  )


def LocationAttributeConfig(prompt_func=location_util.PromptForLocation):
  return concepts.ResourceParameterAttributeConfig(
      name='location',
      help_text='The Cloud location for the {resource}.',
      fallthroughs=[
          deps.ArgFallthrough('--location'),
          deps.Fallthrough(
              function=prompt_func,
              hint='choose one from the prompted list of available locations',
          ),
      ],
  )


def GetLocationResourceSpec():
  return concepts.ResourceSpec(
      'faultinjectiontesting.projects.locations',
      resource_name='location',
      locationsId=LocationAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
  )


def GetLocationResourceArg(
    arg_name='location',
    help_text=None,
    positional=False,
    required=False,
):
  """Constructs and returns the Location Resource Argument."""

  help_text = help_text or 'Location'

  return concept_parsers.ConceptParser.ForResource(
      '{}{}'.format('' if positional else '--', arg_name),
      GetLocationResourceSpec(),
      help_text,
      required=required,
  )


def AddDescribeFaultFlags(parser):
  GetFaultResourceArg().AddToParser(parser)


def AddDeleteFaultFlags(parser):
  GetFaultResourceArg().AddToParser(parser)


def AddCreateFaultFlags(parser):
  GetFaultResourceArg().AddToParser(parser)
  parser.add_argument(
      '--file',
      type=arg_parsers.YAMLFileContents(),
      metavar='FILE',
      help='The file containing the fault to be created.',
      required=True,
  )


def AddUpdateFaultFlags(parser):
  GetFaultResourceArg().AddToParser(parser)
  parser.add_argument(
      '--file',
      type=arg_parsers.YAMLFileContents(),
      metavar='FILE',
      help='The file containing the updated fault config.',
      required=True,
  )


def AddListFaultFlags(parser):
  GetLocationResourceArg().AddToParser(parser)
  parser.add_argument(
      '--service-name',
      type=str,
      help='service name.',
      required=False,
  )
  parser.add_argument(
      '--experiment-name',
      type=str,
      help='experiment name.',
      required=False,
  )


def AddDescribeExperimentFlags(parser):
  GetExperimentResourceArg().AddToParser(parser)


def AddDeleteExperimentFlags(parser):
  GetExperimentResourceArg().AddToParser(parser)


def AddCreateExperimentFlags(parser):
  GetExperimentResourceArg().AddToParser(parser)
  parser.add_argument(
      '--file',
      type=arg_parsers.YAMLFileContents(),
      metavar='FILE',
      help='The file containing the experiment to be created.',
      required=True,
  )


def AddUpdateExperimentFlags(parser):
  GetExperimentResourceArg().AddToParser(parser)
  parser.add_argument(
      '--file',
      type=arg_parsers.YAMLFileContents(),
      metavar='FILE',
      help='The file containing the updated experiment config.',
      required=True,
  )


def AddListExperimentFlags(parser):
  GetLocationResourceArg().AddToParser(parser)
  parser.add_argument(
      '--sd-service-name',
      type=str,
      help='service name.',
      required=False,
  )


def AddDescribeJobFlags(parser):
  GetJobResourceArg().AddToParser(parser)


def AddListJobFlags(parser):
  """Add job list Flags."""

  GetLocationResourceArg().AddToParser(parser)
  concept_parsers.ConceptParser([
      presentation_specs.ResourcePresentationSpec(
          '--experiment',
          GetExperimentResourceSpec(),
          'The experiment resource.',
          prefixes=True,
          required=False,
      ),
      presentation_specs.ResourcePresentationSpec(
          '--fault',
          GetFaultResourceSpec(),
          'The fault resource.',
          prefixes=True,
          required=False,
      ),
  ]).AddToParser(parser)
  parser.add_argument(
      '--target-name',
      type=str,
      help='target name.',
      required=False,
  )
  parser.add_argument(
      '--days',
      type=int,
      help='Days.',
      required=False,
  )


def AddCreateJobFlags(parser):
  """Add job Create Flags."""

  concept_parsers.ConceptParser(
      [
          presentation_specs.ResourcePresentationSpec(
              'JOB',
              GetJobResourceSpec(),
              'The Job resource.',
              flag_name_overrides={'location': '--location'},
              prefixes=True,
              required=True,
          ),
          presentation_specs.ResourcePresentationSpec(
              '--experiment',
              GetExperimentResourceSpec(),
              'The experiment resource.',
              # This hides the location flag for experiment resource.
              flag_name_overrides={
                  'location': '',
                  'project': '',
              },
              prefixes=True,
              required=True,
          ),
      ],
  ).AddToParser(parser)
  parser.add_argument(
      '--fault-targets',
      type=arg_parsers.ArgList(),
      metavar='LIST',
      help='targets for the faults in experiment. Provide them in sequence',
      required=True,
  )
  parser.add_argument(
      '--dry-run',
      action='store_true',
      default=False,
      help='Dry run mode.',

  )


def AddDeleteJobFlags(parser):
  GetJobResourceArg().AddToParser(parser)
